# CI rules for internally hosted FSL projects


This project hosts Gitlab CI automation routines for internally hosted FSL
projects. These routines are hosted at
https://git.fmrib.ox.ac.uk/fsl/fsl-ci-rules, and are applied to most FSL
project repositories.


> **Note:** Currently, all implemented rules are related to the building and
> deployment of conda packages for FSL projects, however there is scope for
> adding more general rules (e.g. testing, building/deploying documentation,
> etc).


Every FSL project comprises **two** git repositories:


 * The _project_ repository contains the project source code/resources, and is
   typically hosted at `https://git.fmrib.ox.ac.uk/fsl/<project>`.
 * The _recipe_ repository contains a conda recipe for the project, and is
   typically hosted at `https://git.fmrib.ox.ac.uk/fsl/conda/fsl-<project>`.


_This_ repository contains CI rules which are applied to both project and
recipe repositories to facilitate automatic building and deployment of FSL
conda packages.


**Important:** The CI rules are written under the assumption that the
following conventions hold. If these conventions are not followed, some rules
will not function correctly:

 - The FSL **project** repository is hosted at
   `https://git.fmrib.ox.ac.uk/fsl/<project>`.
 -  The accompanying conda **recipe** repository is hosted at
   `https://git.fmrib.ox.ac.uk/fsl/conda/fsl-<project>`, with with `<project>`
   all lower-case<sup>*</sup>.
 - The conda package name (as defined in the recipe `meta.yaml` file) is
   `fsl-<project>`, with `<project>` all lower-case.
 - The `master` branch of a FSL project repository is used as the current
   internal/staging release.
 - New stable releases are created as **tags** on the `master` branch of the
   FSL **project** repository. A tag is used as the version number for a
   project, and must comprise a sequence of monotonically increasing integers,
   separated with dots, and with an optional leading `v`.

> <sup>*</sup> Conda package names are required to be all lower-case.


## Overview


When commits are merged/pushed to the master branch of a FSL project
repository at `https://git.fmrib.ox.ac.uk/fsl/<project>`, the conda recipe for
that project (typically hosted at
`https://git.fmrib.ox.ac.uk/fsl/conda/fsl-<project>`) is automatically
re-built against the new state of the project. The resulting conda package is
uploaded to an internal _staging_ conda channel, which is used for
internal/test FSL releases.


When a new release of a FSL project is tagged on the project repository, the
conda recipe for that project is automatically updated and re-built against
the tagged version. The resulting conda package is uploaded to a _production_
conda channel, which forms the basis for public FSL releases.


> **Note:** Both upload steps must be triggered manually via the Gitlab web
> interface by a project maintainer.


Specifically, this repository implements CI rules to perform the following
actions on the project and recipe repositories of a FSL project:

 - Automatic building of a "test" conda package every time commits are
   pushed/merged into _any_ branch of the **project** repository, and
   uploading to a staging channel when commits are pushed/merged into the
   _master_ branch.

 - Automatic opening of merge requests on the **recipe** repository every time
   a new tag is aded to the **project** repository, which will update the
   project version number in the conda recipe.

 - Automatic building of a "stable" conda package every time commits are
   pushed/merged into the _master_ branch of the **recipe** repository, and
   uploading to a production channel.

 - Automatic checking of downstream (dependant) project compatibility before a
   new package is uploaded to the staging or production conda channels. For
   each project which is dependent on the updated package, a test build is
   performed. If a build fails, an issue is automatically opened on the
   **recipe** repository of the downstream project.


### Organisation

This repository is organised as follows:

 - The `rules` directory contains YAML files which specify all of the CI jobs
   that are executed - these are included by the main `.gitlab-ci.yml` file
   (which itself just contains some initialisation logic common to all jobs).
 - The `scripts` directory contains shell and python scripts which are called
   by the CI jobs.
 - The `utils` directory contains some python modules used by the `scripts`,
   along with some python scripts that are intended to be manually invoked -
   see below.


## Bootstrapping a new project


A handful of Python scripts are present in the `utils` sub-directory which can
be used to bootstrap CI and conda integration for new or un-configured
projects.  These scripts are intended to be executed manually by a FSL
maintainer - they should only need to be executed once. The scripts require
Python 3.7, `pyyaml`, `jinja2`, and `networkx`, and the
`create_conda_recipe.py` script works best if run in an environment with an
active FSL installation, so using `fslpython` is a good option.


> **Note:** All of the steps which these scripts perform can be performed
> manually via the gitlab web interface, but performing them in this manner
> becomes very tedious after the third or fourth time.


 - `utils/create_conda_recipe.py`: Creates an initial conda recipe
   repository for a FSL project repository from a standard template (see
   `utils/templates/`).
 - `utils/configure_repositories.py`: Automatically configure a project
   and recipe repository to run the `fsl-ci-rules`.
 - `utils/set_gitlab_variables.py`: Automatically set variables on one
   or more gitlab repositories. This script does not usually need to be used,
   but can be useful for debugging.


### Example


> **Note:** The steps below are a suggestion. If you are comfortable working
> with git repositories, conda environments, and building conda packages by
> hand, see the help documentation for the `create_conda_recipe.py` script,
> and feel free to work in whichever way you see fit.


Let's say we have developed a new FSL project at
`https://git.fmrib.ox.ac.uk/fsl/my_cool_project`. To integrate our new
project into the FSL conda and CI ecosystem, we can follow these steps:


1. Make sure that you have the latest FSL installed. This is not strictly
   necessary, but will give you a suitable Python environment to use, and will
   allow the dependencies of your project to be identified more accurately.

2. If you haven't already done so, create a Gitlab API access token with
   read+write ("`api`") access. This can be done via the Gitlab web interface
   via _User settings -> Access Tokens_.

3. Clone **this** repository:

        git clone https://git.fmrib.ox.ac.uk/fsl/fsl-ci-rules
        cd fsl-ci-rules

4. Create a conda recipe repository for the new project - this step will
   create a recipe repository at
   `https://git.fmrib.ox.ac.uk/fsl/conda/fsl-my_cool_project`:

        fslpython utils/create_conda_recipe.py -t <token> \
            fsl/my_cool_project

5. Configure the project and recipe repositories - this will enable
   integration with the `fsl-ci-rules` CI configuration on both repositories:


        fslpython utils/configure_repositories.py -t <token> \
            fsl/my_cool_project \
            fsl/conda/fsl-my_cool_project

6. To test that the CI integration is working:
   - Trigger a pipeline on the master branch of the `fsl/my_cool_project`
     repository via the Gitlab web interface (_CI / CD -> Run Pipeline_).
   - Create a tag on the `fsl/my_cool_project` repository.

7. Manually edit the recipe at the `fsl/conda/fsl-my_cool_project` repository
   until your conda package builds successfully. You will probably have to
   adjust the requirements of your project, as the routines used to
   automatically identify requirements are not perfect.


## Configuration


In order to interact with Gitlab projects, a Gitlab API access token, which
has full (read+write) access to the Gitlab API, must be created. Access tokens
can be created via the Gitlab web interface. This token will be stored in a
variable on all repositories - see the steps below.


> **Note:** The remaining steps can be performed automatically with the
> `utils/configure_repositories.py` script mentioned above.  Alternately,
> they can be performed via the Gitlab web interface.


On **both project and recipe repositories** you must enable at least one
Gitlab CI runner:

 - One runner which is installed on the server that hosts the conda channel
   directories, with the `docker` and `fslconda_channel_host` tags. This
   runner will be used to run all CI jobs, including those that deploy built
   conda packages. More details on how this runner needs to be configured can
   be found below.

 - One or more runners with the `docker` tag. These runners will be used to
   execute any CI jobs that do not involve deploying conda packages, including
   building `noarch` packages, and binary packages for Linux.

 - One or more runners with the `macOS` tag. These runners will be used to
   build binary packages for macOS. An assumption is made that `conda` is
   installed and available in the environment provided by these runners.

Next, on **both project and recipe repositories**, for a FSL project which does
not require any custom CI rules of its own (most FSL projects fall into this
category), all that needs to be done is to set the *Custom CI configuration
path* option to refer to this file, by giving it the value:

    .gitlab-ci.yml@fsl/fsl-ci-rules

This option can be found under _Settings -> CI / CD -> General Pipelines_,
in the Gitlab web interface for each project.

Finally, you need to set the following variables in the _Settings -> CI / CD ->
Variables_ section of the Gitlab web interface. More detail on these variables
can be found in the **Configuration** section, below:

 - `FSL_CI_API_TOKEN`: Required on both **project** and **recipe**
   repositories. Gitlab API token which is used by CI jobs for read+write
   access via the Gitlab REST API.

 - `FSLCONDA_RECIPE`: Required on the **recipe** repository. A flag which is
   used to distinguish project repositories from recipe repsitories. The value
   is not checked - only whether or not the variable is set.

 - `FSLCONDA_RECIPE_URL`: Required on the **project** repository, but only
   when its accompanying recipe repository is not named according to the
   `fsl/conda/fsl-<project>` convention. Contains the full URL to the conda
   recipe repository.

The above instructions should suffice for most FSL projects. However, for a FSL
project which already has its own `.gitlab-ci.yml` configuration file, instead
of setting the *Custom CI configuration path* setting on the **project
repository**:

 1. The `.gitlab-ci.yml` file of the project should include the `.gitlab-ci.yml`
    file of this repository via the following section:

        include:
         - project: fsl/fsl-ci-rules
           file:    .gitlab-ci.yml

 2. `fsl-ci-pre` `fsl-ci-build` and `fsl-ci-deploy` should be added to the CI
    stages, for example:

        stages:
          - build
          - test
          - fsl-ci-pre
          - fsl-ci-build
          - fsl-ci-deploy


### Setting the FSL conda channel URLs


In order to facilitate automatic building of conda packages, the following
steps need to be performed.

  1. In `.gitlab-ci.yml` of *this* repository, set the following variables. It
     is assumed that the same values will be used for all FSL projects:

      - `FSLCONDA_STAGING_CHANNEL_URL`: `https://` URL of the FSL conda
        staging channel, allowing built conda packages of development/testing
        releases to be downloaded.

      - `FSLCONDA_PRODUCTION_CHANNEL_URL`: `https://` URL of the FSL conda
        staging channel, containing built conda packages of stable/production
        releases.


### Configuring the channel host Gitlab runner


In order to facilitate automatic deployment of built conda packages:

 1. The Gitlab CI runner which is used to run the deployment jobs must be
    running on a server which has access to the channel directories.
 2. The runner must have the `docker` and `fslconda_channel_host` tags.
 3. The runner must be configured so that it has read+write access to the
    channel directories by way of docker volumes mountaed (within the
    container) at `/fslconda_staging` and `/fslconda_production`. This can be
    achieved in the runner `config.toml` file by including something like the
    following:

        [runners.docker]
          volumes = ["/host/path/to/staging_channel:/fslconda_staging:rw",
                     "/host/path/to/production_channel:/fslconda_production:rw"]
