#!/usr/bin/env python
#
# Build a conda package from a FSL recipe repository.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
# (See rules/fsl-ci-conda-rules.yml)


import            os
import os.path as op
import            sys

from fsl_ci_utils       import sprun, tempdir, fprint
from fsl_ci_utils.conda import load_meta_yaml


def pre_build_recipe(url, ref, output_dir, *channels):
    """Checks out and builds the recipe at the given URL,
    storing the built package in the given destination.
    """

    with tempdir():
        fprint(f'Pre-building recipe from {url} (ref: {ref})')
        sprun(f'git clone {url} recipe')
        build_recipe('recipe', ref, output_dir, *channels)


def build_recipe(recipe_dir, ref, output_dir, *channels):
    """Build the conda recipe in the given directory. """

    env = dict(os.environ)

    if ref == '':
        env.pop('FSLCONDA_REVISION', None)
    else:
        env['FSLCONDA_REVISION'] = ref

    cmd = 'conda build -c conda-forge -c defaults'
    for chan in channels:
        cmd = f'{cmd} -c {chan}'

    sprun(f'{cmd} --output-folder={output_dir} {recipe_dir}', env=env)


def get_platform_shortcut_if_not_applicable(pkgname, jobname):
    """The FSL ci conda rules are organised such that a linux, macOS, and
    noarch job is started for every package build. This function checks
    the package type, and the job platform, and exits early if the job
    platform is not applicable for the package type.

    If the job is not short-cutted, the platform ("noarch", "linux" or
    "macos") is returned.
    """
    meta   = load_meta_yaml('meta.yaml')
    noarch = ('build' in meta) and ('noarch' in meta['build'])

    # job name is one of
    #  - build-linux-conda-package
    #  - build-macos-conda-package
    #  - build-noarch-conda-package
    platform = jobname.split('-')[1]

    if noarch and platform != 'noarch':
        print(f'{pkgname} is a noarch package, and does not need to be '
              f'built for {platform} (job: {jobname}). Aborting build.')
        sys.exit(0)
    if (not noarch) and platform == 'noarch':
        print(f'{pkgname} is a binary package, and cannot be built as a '
              f'noarch package (job: {jobname}). Aborting build.')
        sys.exit(0)

    return platform


def main():
    """Build a conda package from a FSL recipe repository.

    This script assumes that the following environment variables
    are set, in addition to the Gitlab CI predefined environment variables:

     - PROJECT_REF: Name of the git ref (e.g. tag, branch) to build the recipe
       from. If empty, ref specified in the recipe meta.yaml file is used.

     - STAGING: An empty or non-empty value to indicate which FSL conda
       channel URL to download dependencies from.  If empty, the
       FSLCONDA_PRODUCTION_CHANNEL_URL is used, otherwise the
       FSLCONDA_STAGING_CHANNEL_URL is used.

     - PREBUILD: A string containing one or more recipe repositories, which
       should be pre-built and installed as dependencies of the package being
       built. This string must have the form:

         "ref1:repo1 ref2:repo2 .."

       i.e. each repository comprises a git reference (e.g.  tag, branch),
       followed by a colon, followed by the URL to the git repository.
       Multiple repositories can be specified by separating them with spaces.
    """

    project_ref   = os.environ['PROJECT_REF']
    recipe_name   = os.environ['CI_PROJECT_NAME']
    recipe_url    = os.environ['CI_PROJECT_URL']
    recipe_ref    = os.environ['CI_COMMIT_REF_NAME']
    job_name      = os.environ['CI_JOB_NAME']
    staging       = os.environ['STAGING']
    prebuild      = os.environ['PREBUILD']
    prebuild      = [p.split(':', 1) for p in prebuild.split()]
    local_channel = './local_channel'
    platform      = get_platform_shortcut_if_not_applicable(recipe_name,
                                                            job_name)

    # conda has some seriously fucked up behaviour -
    # if you try to direct the build into a directory
    # called "noarch", it thinks that you are directing
    # it to a channel subdirectory, and explodes. So
    # we output to _platform_, rather thah platform.
    output_dir    = op.join(os.getcwd(), 'conda_build', f'_{platform}_')

    if staging == 'true':
        channel_url = os.environ['FSLCONDA_STAGING_CHANNEL_URL']
    else:
        channel_url = os.environ['FSLCONDA_PRODUCTION_CHANNEL_URL']

    os.makedirs(output_dir)
    os.mkdir(local_channel)

    # pre-build local packages, make them
    # available in a local channel directory
    for ref, url in prebuild:
        pre_build_recipe(url, ref, local_channel, channel_url)
    sprun(f'conda index {local_channel}')

    fprint('************************************')
    fprint(f'Building conda recipe for:               {recipe_name}')
    fprint(f'Recipe URL:                              {recipe_url}')
    fprint(f'Recipe revision:                         {recipe_ref}')
    fprint(f'Pre-built packages:                      {prebuild}')
    fprint( 'Project revision (empty means to ')
    fprint(f'  build release specified in meta.yaml): {project_ref}')
    fprint(f'FSL conda channel URL:                   {channel_url}')
    fprint('************************************')
    build_recipe('.', project_ref, output_dir, local_channel, channel_url)


if __name__ == '__main__':
    main()
