#!/usr/bin/env python
#
# This script is executed on a FSL recipe repository.  All "downstream"
# packages (those which are dependent on this package) are identified, and a
# CI pipeline is triggered on each of them to perform a test build.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os
import sys

from fsl_ci_utils.conda     import (load_meta_yaml,
                                    read_channel_repodata,
                                    load_packages,
                                    build_dependency_graph)
from fsl_ci_utils.conda_api import (trigger_pipeline,
                                    wait_on_pipeline)


def trigger_pipeline_on_packages(upstream, packages, server, token, staging):
    """For each package, looks up the gitlab repository URL for each, and
    then triggers a pipeline on its master branch.
    """

    pipelines = []

    # these variables will be passed to the
    # build-conda-recipe job on the downstream
    # recipe repository:
    #
    #  - PROJECT_REF:      Build recipe from the project repository
    #                      master branch, or from the ref specified
    #                      in the recipe meta.yaml
    #  - PREBUIlD:         Prebuild this recipe locally, rather
    #                      than installing it from the channel.
    #  - STAGING:          whether to build from staging or
    #                      production channels
    #  - DEPLOY:           Built package is *not* for deployment
    #  - CHECK_DOWNSTREAM: Do *not* perform downstream check -
    #                      we only check direct dependants, and
    #                      do not recursively run
    #                      trigger-check-downstrem-packages all
    #                      the way down the dependency tree.
    ref       = 'master' if (staging == 'true') else ''
    variables = {
        'PROJECT_REF'      : ref,
        'PREBUILD'         : f'{ref}:{upstream}',
        'STAGING'          : staging,
        'DEPLOY'           : '',
        'CHECK_DOWNSTREAM' : '',
    }

    for pkg in packages:
        # we assume that the recipe of a conda package with name <package>
        # has a recipe repository located at <server>/fsl/conda/<package>
        recipe_path = f'fsl/conda/{pkg}'

        print(f'Triggering pipeline on downstream package {recipe_path}')

        pipelines.append((recipe_path, trigger_pipeline(
            recipe_path, 'master', server, token, variables)))

    passed = True

    for recipe_path, pipeline in pipelines:
        status = wait_on_pipeline(
            recipe_path, pipeline['id'], server, token)
        passed = passed and (status == 'success')

    return passed


def main():

    staging    = os.environ['STAGING']
    gitlab_url = os.environ['CI_SERVER_URL']
    token      = os.environ['FSL_CI_API_TOKEN']
    upstream   = os.environ['CI_PROJECT_URL']

    if staging == 'true':
        channel_url = os.environ['FSLCONDA_STAGING_CHANNEL_URL']
    else:
        channel_url = os.environ['FSLCONDA_PRODUCTION_CHANNEL_URL']

    # load package information from the
    # channel, and build a dependency graph
    # from it
    meta  = read_channel_repodata(channel_url)
    pkgs  = load_packages(*meta)
    graph = build_dependency_graph(pkgs)

    # identify the packages which directly
    # depend on the updated package, and
    # trigger a rebuild pipeline on their
    # recipe repositories.
    pkgname    = load_meta_yaml('meta.yaml')['package']['name']
    downstream = list(graph.reverse().neighbors(pkgname))

    if len(downstream) == 0:
        print(f'{pkgname} nas no downstream dependants - nothing to do!')
    else:
        result = trigger_pipeline_on_packages(
            upstream, downstream, gitlab_url, token, staging)
        sys.exit(0 if result else 1)



if __name__ == '__main__':
    main()
