#!/usr/bin/env python
#
# This script is run from a CI job on the project repository. It triggers
# a pipeline on the accompanying recipe repository
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import os
import sys

from fsl_ci_utils.conda     import get_recipe_url
from fsl_ci_utils.conda_api import (trigger_pipeline,
                                    wait_on_pipeline)



def main():

    server       = os.environ['CI_SERVER_URL']
    token        = os.environ['FSL_CI_API_TOKEN']
    project_path = os.environ['CI_PROJECT_PATH']
    project_ref  = os.environ['CI_COMMIT_REF_NAME']
    deploy       = os.environ.get('DEPLOY', None)
    recipe       = get_recipe_url(project_path, server)['path']

    if deploy is None:
        deploy = 'true' if (project_ref == 'master') else ''

    # These variables are passed to the
    # build-conda-package job on the
    # recipe repository. They control
    # how the package is built, whether
    # a downstream compatibility check
    # is performed, and whether the
    # built package is deployed to a
    # conda channel.
    variables   = {
        'PROJECT_REF'      : project_ref,
        'STAGING'          : 'true',
        'CHECK_DOWNSTREAM' : 'true' if (project_ref == 'master') else '',
        'DEPLOY'           : deploy
    }

    # start the pipeline, wait for it to finish
    pipeline = trigger_pipeline(recipe, 'master', server, token, variables)
    pid      = pipeline['id']
    url      = pipeline['web_url']

    print(f'Triggered pipeline on {recipe} - see {url} for more details.')

    status = wait_on_pipeline(recipe, pid, server, token)

    if status in ('failed', 'cancelled', 'skipped'):
        print(f'Pipeline {status}! Check {url} for details.')
        sys.exit(1)
    if status in ('success', 'manual'):
        print(f'Pipeline succeeded! Check {url} for details.')
        sys.exit(0)
    if status in ('manual'):
        print('Pipeline needs manual intervention to proceed - '
              f'check {url} for details.')
        sys.exit(0)


if __name__ == '__main__':
    main()
