#!/usr/bin/env python
#
# This script is executed when a new tag is added to a FSL project
# repository. It opens a merge request on a conda recipe repository to replace
# the project version number with the new tag, in the recipe meta.yaml file.
#
# Thanks to: https://gitlab.com/tmaier/gitlab-auto-merge-request
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import                    os
import                    re
import                    sys

from fsl_ci_utils           import  tempdir, sprun, is_valid_project_version
from fsl_ci_utils.conda     import  get_recipe_url
from fsl_ci_utils.conda_api import (http_request,
                                    lookup_project_id,
                                    open_merge_request)


GITLAB_URL      = os.environ['CI_SERVER_URL']
APIURL          = f'{GITLAB_URL}/api/v4/'
TOKEN           = os.environ['FSL_CI_API_TOKEN']
PROJECT_PATH    = os.environ['CI_PROJECT_PATH']
PROJECT_NAME    = os.environ['CI_PROJECT_NAME']
PROJECT_VERSION = os.environ['CI_COMMIT_REF_NAME']


COMMIT_MESSAGE = f"""
MNT: Update {PROJECT_NAME} version number
to {PROJECT_VERSION} (automatic fsl-ci-rules)
""".strip()


MERGE_REQUEST_MESSAGE = f"""
This merge request was triggered by a new tag being added to
the {PROJECT_PATH} project.

Accepting this merge request will cause the conda recipe for
{PROJECT_NAME} to be updated to version {PROJECT_VERSION}.

If all goes well, a new conda package will built for
{PROJECT_NAME} ({PROJECT_VERSION}), and will be uploaded to
the FSL production conda channel.

(MR automatically generated by fsl-ci-rules running on the
{PROJECT_PATH} repository).
""".strip()


def check_mr_already_open(project_id, source_branch, target_branch):
    """Checks to see if a MR is already open on the specified project, from
    the source to target     """

    url      = f'{APIURL}/merge_requests?state=opened'
    response = http_request(url, TOKEN)

    for mr in response:
        if all((mr['source_project_id'] == project_id,
                mr['target_project_id'] == project_id,
                mr['source_branch']     == source_branch,
                mr['target_branch']     == target_branch)):
            return True

    return False


def patch_recipe(metayaml, newversion):
    """Search for a line in the recipe meta.yaml which
    has the structure:

    {% set version = '<oldversion>' %}

    and replace <oldversion> with newversion
    """

    pat   = r"^{% *set version( *)= *'(.*)' *%} *$"
    meta  = list(open(metayaml).read().split('\n'))
    match = None

    for i, line in enumerate(meta):
        match = re.fullmatch(pat, line)
        if match:
            space      = match.group(1)
            oldversion = match.group(2)
            meta[i]    = f'{{% set version{space}= \'{newversion}\' %}}'
            break
    else:
        raise RuntimeError('Unable to automatically patch meta.yaml!')

    if oldversion == newversion:
        raise RuntimeError('Conda recipe is already referring to '
                           f'the latest version ({newversion})!')

    with open(metayaml, 'wt') as f:
        f.write('\n'.join(meta))


def checkout_and_patch_recipe(url, branch, project_version):

    with tempdir():
        sprun(f'git clone {url} recipe')
        os.chdir('recipe')
        sprun(f'git checkout -b {branch} master')
        patch_recipe('meta.yaml', project_version)
        sprun( 'git add meta.yaml')
        sprun(f'git commit -m "{COMMIT_MESSAGE}"')
        sprun(f'git push origin {branch}')


def main():
    recipe       = get_recipe_url(GITLAB_URL, PROJECT_NAME)
    recipe['id'] = lookup_project_id(recipe['path'], GITLAB_URL, TOKEN)
    branch       = f'mnt/rel-{PROJECT_VERSION}'

    if not is_valid_project_version(PROJECT_VERSION):
        print(f'{PROJECT_VERSION} is not a valid FSL version - aborting!')
        sys,exit(1)

    if check_mr_already_open(recipe['id'], branch, 'master'):
        print(f'A merge request from branch {branch} is '
              f'already open on {recipe["url"]} - aborting!')
        sys.exit(1)

    checkout_and_patch_recipe(recipe["url"], branch, PROJECT_VERSION)

    open_merge_request(recipe["path"],
                       branch,
                       MERGE_REQUEST_MESSAGE,
                       GITLAB_URL,
                       TOKEN)


if __name__ == '__main__':
    main()
