#!/usr/bin/env python
#
# This script can be used to create an initial conda recipe repository
# for a new FSL project, or update an existing recipe repository.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os.path  as op
import textwrap as tw
import             os
import             sys
import             glob
import             shutil
import             argparse

import jinja2   as j2

from fsl_ci_utils           import (indir,
                                    tempdir,
                                    sprun)
from fsl_ci_utils.conda_api import (http_request,
                                    project_exists,
                                    list_project_branches,
                                    create_repository,
                                    get_project_version,
                                    open_merge_request)
from fsl_ci_utils.conda     import  gen_recipe_path

import fsl_project_dependencies as fsldeps


SERVER_URL = 'https://git.fmrib.ox.ac.uk'
"""Default gitlab instance URL, if not specified on the command.line. """


TEMPLATE_DIR = op.abspath(op.join(op.dirname(__file__), 'templates'))
"""Directory which contains conda recipe template files. """


MERGE_REQUEST_MSG = """
Update conda recipe to latest template-generated recipe. This merge request
was automatically generated by fsl-ci-rules/utils/create_conda_recipe.py.
""".strip()


VERSION_OVERRIDE = {
    'libgd'     : '2.2.5',
    'boost-cpp' : '1.67',
    'libnlopt'  : '2.6.2',
    'libpng'    : '1.6.37',
    'libsqlite' : '3.33.0',
    'libxml2'   : '2.9.10',
    'libxmlpp'  : '2.40.1',
    'tk'        : '8.6.10',
    'zlib'      : '1.2.11',
    'CiftiLib'  : '1.2',
}
"""Some hard-coded version numbers for external ("extras") FSL dependencies.
"""


def project_type(project_dir):
    """Return:
      - "cpp" if the project looks like a conventional C/C++ Makefile-based
        project
      - "python" if the project looks like a Python based project
      - "other" if the project looks like a non-C/C++ Makefile based project
      - "unwknown" if the project doesn't look like anything we can work with.
    """

    if op.exists(op.join(project_dir, 'setup.py')):
        return 'python'

    elif op.exists(op.join(project_dir, 'Makefile')):
        for pattern in ['*.cc', '*.h', '*.cpp', '*.cxx', '*.c']:
            if glob.glob(op.join(project_dir, '**', pattern), recursive=True):
                return 'cpp'
        return 'other'
    else:
        return 'unknown'


def gen_repository_url(project_path, server, token=None):
    """Generates a URL for the given project. """

    if token is not None:
        prefix, suffix = server.split('://')
        server         = f'{prefix}://gitlab-ci-token:{token}@{suffix}'

    return f'{server}/{project_path}.git'


def clone_repository(path, server, token, destdir):
    """Clone the specified gitlab repository into destdir. """
    url = gen_repository_url(path, server, token)
    sprun(f'git clone {url} {destdir}')


def create_local_recipe_repository(recipe_dir,
                                   recipe_path,
                                   server,
                                   token):
    """If recipe_path already exists on gitlab, it is cloned into recipe_dir.
    Otherwise recipe_dir is turned into a git repository with an initial
    commit and remote origin configured.
    """
    recipe_url = gen_repository_url(recipe_path, server, token)

    os.mkdir(recipe_dir)
    with indir(recipe_dir):
        sprun( 'git init .')
        sprun(f'git remote add origin {recipe_url}')


def commit_and_push_recipe_repository(recipe_dir,
                                      project_path,
                                      branch='master',
                                      push=True):
    """Commit the current state of the local recipe repository and push
    it to gitlab.
    """
    with indir(recipe_dir):

        msg = f'MNT: Initial auto-generated conda ' \
              f'recipe for {project_path}'

        sprun(f'git checkout -B {branch}')
        sprun( 'git add *')
        sprun(f'git commit -m "{msg}"')

        if push:
            sprun(f'git push origin {branch}')


def get_fsl_project_dependencies(project_path, project_dir, server, token):
    """Returns a list of all FSL dependencies of the project as (project,
    version) tuples.
    """
    if ('FSLDIR' not in os.environ) or ('FSLDEVDIR' not in os.environ):
        print('FSLDIR/FSLDEVDIR are not set - cannot identify '
              f'dependencies for project {project_path}')
        return []

    deps = fsldeps.get_project_dependencies(project_dir)
    vers = []

    for i, dep in enumerate(deps):
        if dep in VERSION_OVERRIDE:
            dver = VERSION_OVERRIDE[dep]
        else:
            try:
                # assume that the dependency has
                # a repo at "server/fsl/dep"
                dver = get_project_version(f'fsl/{dep}', server, token)
                dep  = gen_recipe_path(dep).rsplit('/')[-1]

            except Exception:
                dver = 'UNKNOWN'

        vers.append(dver)
        deps[i] = dep
    return list(zip(deps, vers))


def create_python_recipe_template(project_path,
                                  recipe_path,
                                  recipe_dir,
                                  server,
                                  token):
    """Creates a conda recipe for the Python project in project_dir, storing
    the recipe files in recipe_dir.
    """
    recipe_name       = recipe_path.rsplit('/')[-1]
    metayaml_template = op.join(TEMPLATE_DIR, 'meta.yaml.python.template')
    metayaml_dest     = op.join(recipe_dir,   'meta.yaml')
    base_version      = get_project_version('fsl/base',   server, token)
    project_version   = get_project_version(project_path, server, token)

    meta              = {
        'template_name'              : recipe_name,
        'template_repository'        : gen_repository_url(project_path, server),
        'template_recipe_repository' : gen_repository_url(recipe_path, server),
        'template_version'           : project_version,
        'fslbaseversion'             : base_version,
    }

    with open(metayaml_template, 'rt') as f:
        template = j2.Template(f.read())

    with open(metayaml_dest, 'wt') as f:
        f.write(template.render(**meta) + '\n')


def create_cpp_recipe_template(project_path,
                               recipe_path,
                               project_dir,
                               recipe_dir,
                               server,
                               token,
                               metayaml_template='meta.yaml.cpp.template'):
    """Creates a conda recipe for the C/C++ project in project_dir, storing the
    recipe files in recipe_dir.
    """

    recipe_name       = recipe_path.rsplit('/')[-1]
    buildsh_template  = op.join(TEMPLATE_DIR, 'build.sh.template')
    metayaml_template = op.join(TEMPLATE_DIR, metayaml_template)
    buildsh_dest      = op.join(recipe_dir,   'build.sh')
    metayaml_dest     = op.join(recipe_dir,   'meta.yaml')
    base_version      = get_project_version('fsl/base',   server, token)
    project_version   = get_project_version(project_path, server, token)
    deps              = get_fsl_project_dependencies(project_path,
                                                     project_dir,
                                                     server,
                                                     token)
    meta = {
        'template_name'              : recipe_name,
        'template_repository'        : gen_repository_url(project_path, server),
        'template_recipe_repository' : gen_repository_url(recipe_path, server),
        'template_version'           : project_version,
        'fslbaseversion'             : base_version,
        'dependencies'               : deps,
    }

    with open(buildsh_template, 'rt') as inf, \
         open(buildsh_dest, 'wt') as outf:
        outf.write(inf.read())

    with open(metayaml_template, 'rt') as f:
        template = j2.Template(f.read())

    with open(metayaml_dest, 'wt') as f:
        f.write(template.render(**meta) + '\n')


def create_other_recipe_template(project_path,
                                 recipe_path,
                                 project_dir,
                                 recipe_dir,
                                 server,
                                 token):
    """Creates a conda recipe for the FSL project in project_dir, storing the
    recipe files in recipe_dir.
    """
    create_cpp_recipe_template(project_path,
                               recipe_path,
                               project_dir,
                               recipe_dir,
                               server,
                               token,
                               'meta.yaml.other.template')


def create_recipe_template(project_path,
                           recipe_path,
                           project_dir,
                           recipe_dir,
                           server,
                           token):
    """Creates a conda recipe template for the given project. """

    ptype = project_type(project_dir)
    if ptype == 'cpp':
        create_cpp_recipe_template(project_path,
                                   recipe_path,
                                   project_dir,
                                   recipe_dir,
                                   server,
                                   token)
    elif ptype == 'python':
        create_python_recipe_template(project_path,
                                      recipe_path,
                                      recipe_dir,
                                      server,
                                      token)
    elif ptype == 'other':
        create_other_recipe_template(project_path,
                                     recipe_path,
                                     project_dir,
                                     recipe_dir,
                                     server,
                                     token)
    else:
        print(f'{project_path} does not look like '
              'a FSL or Python project - aborting.')
        sys.exit(1)


def gen_branch_name(recipe_path, server, token):
    """Generates a unique branch name on an existing recipe repository
    to open a MR with the updated recipe.
    """

    branch      = 'mnt/recipe_update'
    allbranches = list_project_branches(recipe_path, server, token)

    count = 0
    while branch in allbranches:
        count += 1
        branch = f'mnt/recipe_update-{count}'

    return branch


def parseArgs():
    usage  = 'Usage: create_conda_recipe.py -t <token> [options] project_path'
    desc   = tw.dedent("""
    Create or update a conda recipe repository for the specified project.
    """).strip()
    parser = argparse.ArgumentParser(
        usage=usage,
        description=desc,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    helps = {
        'project_path' :
        'Path (namespace/project) of the gitlab project repository',

        'recipe_path' :
        'Path (namespace/project) of the gitlab recipe repository (default: '
        'automatically generated from project path). ',

        'server' :
        f'Gitlab server (default: {SERVER_URL})',

        'token' :
        'Gitlab API access token with read+write access',

        'local' :
        'Do not push recipe repository to gitlab - instead, create a local '
        'repository at the specified path.',
    }

    parser.add_argument('project_path',        help=helps['project_path'])
    parser.add_argument('-r', '--recipe_path', help=helps['recipe_path'])
    parser.add_argument('-l', '--local',       help=helps['local'])
    parser.add_argument('-s', '--server',      help=helps['server'],
                        default=SERVER_URL)
    parser.add_argument('-t', '--token',       help=helps['token'],
                        required=True)

    args = parser.parse_args()

    if args.local is not None:
        args.local = op.abspath(args.local)
    if args.recipe_path is None:
        args.recipe_path = gen_recipe_path(args.project_path)

    return args


def main():
    args = parseArgs()

    with tempdir():
        project_dir   = op.abspath('project')
        recipe_dir    = op.abspath('recipe')
        recipe_exists = project_exists(args.recipe_path,
                                       args.server,
                                       args.token)

        print(f'Cloning {args.project_path}')
        clone_repository(args.project_path,
                         args.server,
                         args.token,
                         project_dir)

        if recipe_exists:
            print(f'Cloning existing recipe repository ({args.recipe_path})')
            clone_repository(args.recipe_path,
                             args.server,
                             args.token,
                             recipe_dir)
        else:
            print('Creating local recipe repository')
            create_local_recipe_repository(recipe_dir,
                                           args.recipe_path,
                                           args.server,
                                           args.token)

        print('(Re-)Generating conda recipe')
        create_recipe_template(args.project_path,
                               args.recipe_path,
                               project_dir,
                               recipe_dir,
                               args.server,
                               args.token)

        if recipe_exists:
            branch = gen_branch_name(args.recipe_path,
                                     args.server,
                                     args.token)
        else:
            branch = 'master'

        if (args.local is None) and (not recipe_exists):
            print(f'Creating gitlab repository ({args.recipe_path})')
            create_repository(args.recipe_path,
                              args.server,
                              args.token)

        print('Commiting recipe and (maybe) pushing')
        commit_and_push_recipe_repository(recipe_dir,
                                          args.recipe_path,
                                          branch=branch,
                                          push=args.local is None)

        if args.local is not None:
            print(f'Copying repository to {args.local}')
            shutil.copytree(recipe_dir, args.local)

        # open MR
        elif recipe_exists:
            print('Opening MR on existing recipe '
                  f'repository ({args.recipe_path})')
            open_merge_request(args.recipe_path,
                               branch,
                               MERGE_REQUEST_MSG,
                               args.server,
                               args.token)


if __name__ == '__main__':
    main()
