#!/usr/bin/env python
#
# Functions for working with FSL conda recipe repositories, and with conda
# channels.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import                 os
import                 re
import                 dataclasses
import urllib.parse as urlparse

from typing import Dict, List, Union, Tuple

import networkx as nx
import jinja2   as j2
import             yaml

from fsl_ci_utils.conda_api import (http_request,)


def gen_recipe_path(project_path):
    """Generates a conda recipe project name from the given project.

    If the project name has one of the following forms (case doesn't matter):
      - "fsl-<project>"
      - "fsl_<project>"
      - "fsl<project>"

    the recipe name will be "fsl-<project>". Otherwise, the recipe name will
    be "fsl-<project_name>". In both cases, the recipe name will be all lower
    case.

    The full recipe path will be "fsl/conda/<recipe_name>"
    """

    # strip namespaced in case
    # fsl/<project> was passed in
    project_name = project_path.rsplit('/')[-1]
    match = re.match(r'^(fsl[_-]?)?(.*)$', project_name, flags=re.IGNORECASE)
    suffix = match.group(2).lower()

    return f'fsl/conda/fsl-{suffix}'


def get_recipe_url(project_name, server) -> Dict[str, str]:
    """Figures out the URL to the conda recipe associated with the specified
    FSL project.
    """


    # Figure out the name of the conda recipe associated with
    # this project. It defaults to "fsl-<project>", but can
    # be overridden via the FSLCONDA_RECIPE_URL environment
    # variable.
    url          = os.environ.get('FSLCONDA_RECIPE_URL')
    project_name = project_name.rsplit('/')[-1]

    if url is None:
        url = f'{server}/fsl/conda/fsl-{project_name}.git'

    path = urlparse.urlparse(url).path.replace('.git', '')[1:]
    name = path.rsplit('/', 1)[-1]

    return {'url' : url, 'name' : name, 'path' : path}



def is_standard_recipe_path(project_path, recipe_path):
    """Returns True if the given recipe path follows the "standard" convention
    of being "fsl/conda/fsl-<project_name>", False otherwise.
    """
    project_name = project_path.rsplit('/')[-1].lower()
    return recipe_path == f'fsl/conda/fsl-{project_name}'


def load_meta_yaml(filename):
    """Load a conda recipe meta.yaml file."""

    def nil_func(*args, **kwargs):
        return "nil"

    env = {
        'os'                 : os,
        'load_setup_py_data' : lambda : {},
        'compiler'           : nil_func,
        'cdt'                : nil_func,
    }

    with open(filename, 'rt') as f:
        template = j2.Template(f.read())

    meta = template.render(**env)
    meta = yaml.load(meta, Loader=yaml.Loader)

    return meta


@dataclasses.dataclass
class Package:

    name : str
    """Package base name."""

    version : str
    """Latest version available (for any platform). We assume that the
    same versions are available for all supported platforms.
    """

    platforms : List[str]
    """List of platforms (e.g. "noarch", "osx-64") for which this package
    is available.
    """

    dependencies : List[Union[str, 'Package']]
    """References to all packages which this package depends on.  Stored
    as a reference to another Package object for other packages hosted
    on the same channel, or a "package [version-constraint]" string for
    packages hosted elsewhere.
    """


def read_channel_repodata(channel_url : str) -> Tuple[Dict, Dict]:
    """Loads channel and platform metadata from the channel. """

    # load channel and platform metadata - the
    # first gives us a list of all packages that
    # are hosted on the channel, and the second
    # gives us the dependencies of each package.
    channeldata  = http_request(f'{channel_url}/channeldata.json')
    platformdata = {}
    for platform in channeldata['subdirs']:
        url = f'{channel_url}/{platform}/repodata.json'
        platformdata[platform] = http_request(url)

    return channeldata, platformdata


def load_packages(channeldata  : Dict,
                  platformdata : Dict) -> Dict[str, Package]:
    """Creates a Package object for every package hosted in the channel. """

    packages = {}

    # The channeldata.json file contains an entry for
    # every package hosted on the channel. We use this
    # to create a Package object for every package in
    # the channel, but without dependencies - we'll do
    # this in a second pass below.
    for name, meta in channeldata['packages'].items():
        version        = meta['version']
        platforms      = meta['subdirs']
        packages[name] = Package(name, version, platforms, [])

    # Load dependency information - this is stored
    # in platform-specific repodata.json files.
    for pkg in packages.values():

        # assume that version/dependencies are
        # identical across all supported platforms
        platform = pkg.platforms[0]

        # repodata contains one entry for every package
        # file hosted in the channel (i.e. there will be
        # multiple files for each package, each for a
        # different version). Find the entry corresponding
        # to the latest version.
        for fmeta in platformdata[platform]['packages'].values():
            if fmeta['name']    == pkg.name and \
               fmeta['version'] == pkg.version:
                break
        else:
            raise RuntimeError(f'Cannot find package {pkg.name} in '
                               f'{platform} sub-directory!')

        # The entry for each package file contains a list
        # of "package [version-constraint]" strings.
        for dep in fmeta['depends']:
            name = dep.split()[0]

            # Externally hosted package - store
            # the full dependency string
            if name not in packages:
                pkg.dependencies.append(dep)

            # Another package hosted in this
            # channel - store a ref to the
            # Package object
            else:
                pkg.dependencies.append(packages[name])

    return packages


def build_dependency_graph(packages : Dict[str, Package]) -> nx.DiGraph:
    """Builds a directed dependency graph from the given collection of
    packages. Externally hosted packages are not included in the graph.
    """

    g = nx.DiGraph()

    # nodes
    for pkg in packages.values():
        g.add_node(pkg.name)

    # edges
    for pkg in packages.values():
        for dep in pkg.dependencies:
            if isinstance(dep, Package):
                g.add_edge(pkg.name, dep.name)

    return g
