#!/usr/bin/env python
#
# Functions for interacting with Gitlab over its HTTP REST API.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import                    json
import                    time
import                    datetime
import base64          as b64
import functools       as ft
import urllib.parse    as urlparse
import urllib.request  as urlrequest

from fsl_ci_utils import is_valid_project_version


def http_request(url, token=None, data=None, method=None):
    """Submit a HTTP request to the given URL. """

    if method is None:
        if data is None: method = 'GET'
        else:            method = 'POST'

    print(f'{method} {url} ...')

    headers = {}

    if token is not None:
        headers['PRIVATE-TOKEN'] = token

    if data is not None:
        headers['Content-Type'] = 'application/json'
        data                    = json.dumps(data).encode('utf-8')

        print(f'    payload: {data}')

    request  = urlrequest.Request(
        url, headers=headers, data=data, method=method)
    response = urlrequest.urlopen(request).read()

    if len(response) == 0:
        return {}
    else:
        return json.loads(response)


@ft.lru_cache()
def lookup_project_id(project_path, server, token):
    """Look up the integer ID of a gitlab project from its fully qualified
    path.
    """
    project_path = urlparse.quote_plus(project_path)
    url          = f'{server}/api/v4/projects/{project_path}'
    return http_request(url, token)['id']


def project_exists(project_path, server, token):
    """Returns true if the given project exists, false othewrise. """
    try:
        lookup_project_id(project_path, server, token)
        return True
    except Exception:
        return False


def lookup_namespace_id(namespace_path, server, token):
    """Look up the integer ID of a gitlab namespace from its fully qualified
    path.
    """
    url        = f'{server}/api/v4/namespaces'
    namespaces = http_request(url, token)

    for n in namespaces:
        if n['path'] == namespace_path:
            return n['id']

    raise ValueError(f'No namespace matching {namespace_path}')


def get_available_runners(project_path, server, token):
    """Returns a list of the IDs of all specific runners
    which are available to be used for the given project,
    and are not already enabled for it.
    """
    # get all specific runners available to the user
    url     = f'{server}/api/v4/runners'
    runners = http_request(url, token)

    # remove inactive ones, and ones that are already
    # enabled
    runners = [r for r in runners if r['online'] and r['active']]
    runners = [r for r in runners if not runner_is_enabled(
        project_path, r['id'], server, token)]


    return [r['id'] for r in runners]


def get_runner_metadata(runner_id, server, token):
    """Returns metadata for the specified runner. """
    url      = f'{server}/api/v4/runners/{runner_id}'
    response = http_request(url, token)
    return response


def get_runner_tags(runner_id, server, token, lower=True):
    """Returns the list of tags for the specified runner. """

    tags = get_runner_metadata(runner_id, server, token)['tag_list']

    if lower:
        tags = [t.lower() for t in tags]

    return tags


def find_suitable_runners(project_path, tags, server, token):
    """Identifies runners with the specified set of tags which
    are available to be used on the given project.
    """

    tags = [t.lower() for t in tags]

    def match(runner_tags):
        return all([t in runner_tags for t in tags])

    rids  = get_available_runners(project_path, server, token)
    rtags = [get_runner_tags(r, server, token) for r in rids]

    return [r for r, t in zip(rids, rtags) if match(t)]


def lookup_project_tags(project_path, server, token):
    """Return the a list of tags for the given project, or an empty list
    if the project has no tags.

    The tags are sorted such that the most recently updated tag is first
    in the list.
    """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/repository/tags'
    tags = http_request(url, token)
    tags = [t['name'] for t in tags]

    return tags


def get_project_version(project_path, server, token):
    """Return the most recent version of the specified project, returning
    "YYMM.0" if the project does not have any tags.
    """
    tags    = lookup_project_tags(project_path, server, token)
    version = None

    if len(tags) >= 0 and is_valid_project_version(tags[0]):
        version = tags[0]

    if version is None:
        version = datetime.date.today().strftime('%y%m')
        version = f'{version}.0'

    return version


def list_project_branches(project_path, server, token):
    """Returns a list of all branches of the project. """
    pid      = lookup_project_id(project_path, server, token)
    url      = f'{server}/api/v4/projects/{pid}/repository/branches'
    response = http_request(url, token)
    return [r['name'] for r in response]


def get_project_metadata(project_path, server, token):
    """Returns metadata for the specifie project."""
    pid = lookup_project_id(project_path, server, token)
    url = f'{server}/api/v4/projects/{pid}/'
    return http_request(url, token)


def set_project_metadata(project_path, server, token, data):
    """Sets metadata for the specifie project."""
    pid = lookup_project_id(project_path, server, token)
    url = f'{server}/api/v4/projects/{pid}/'
    http_request(url, token, data, method='PUT')

def get_protected_branches(project_path, server, token):
    """Returns a list of all protected branches for the given project. """
    pid      = lookup_project_id(project_path, server, token)
    url      = f'{server}/api/v4/projects/{pid}/protected_branches'
    response = http_request(url, token)
    return [r['name'] for r in response]


def protect_branch(project_path, branch, server, token):
    """Protects the specified branch, so that commits cannot be pushed
    directly to it.
    """
    pid = lookup_project_id(project_path, server, token)

    # GItlab does not let you change the settings
    # on a branch that is already protected, so
    # we first have to unprotect the branch.
    if branch in get_protected_branches(project_path, server, token):
        url = f'{server}/api/v4/projects/{pid}/protected_branches/{branch}'
        http_request(url, token, method='DELETE')

    url  = f'{server}/api/v4/projects/{pid}/protected_branches'
    data = {
        'name'                   : branch,
        'push_access_level'      : '0',
        'merge_access_level'     : '40',
        'unprotect_access_level' : '40',
    }
    http_request(url, token, data=data, method='POST')


def create_repository(project_path, server, token):
    """Create a repository on gitlab. """
    namespace, name = project_path.rsplit('/', 1)
    namespace       = lookup_namespace_id(namespace, server, token)
    url             = f'{server}/api/v4/projects'
    data            = {
        'name'         : name,
        'visibility'   : 'internal',
        'namespace_id' : namespace
    }
    http_request(url, token, data)


@ft.lru_cache()
def download_file(
        project_path, filename, server, token, ref='master', text=True):
    """Download the specified from the specified branch/ref of the project. """
    pid      = lookup_project_id(project_path, server, token)
    filename = urlparse.quote_plus(filename)
    url      = f'{server}/api/v4/projects/{pid}/repository/files/'
    url      = f'{url}/{filename}?ref={ref}'
    contents = http_request(url, token)['content']

    # contents are base64 encoded
    contents = b64.b64decode(contents)

    # bytes or str
    if text:
        contents = contents.decode('utf-8')

    return contents


def update_file(project_path,
                filename,
                contents,
                message,
                server,
                token,
                branch='master'):
    """Update a file on the specified branch of the project. """

    pid      = lookup_project_id(project_path, server, token)
    filename =  urlparse.quote_plus(filename)
    url      = f'{server}/api/v4/projects/{pid}/repository/files/{filename}'
    data     = {
        'filename'       : urlparse.quote_plus(filename),
        'content'        : contents,
        'branch'         : branch,
        'commit_message' : message
    }
    http_request(url,  token, data=data, method='PUT')


def runner_is_enabled(project_path, runner_id, server, token):
    """Return True if the given runner is enabled for the given project,
    False otherwise.
    """
    pid     = lookup_project_id(project_path, server, token)
    url     = f'{server}/api/v4/projects/{pid}/runners'
    runners = http_request(url, token)
    rids    = [r['id'] for r in runners]

    return runner_id in rids


def enable_runner(project_path, runner_id, server, token):
    """Enables the specified runner on the specified project."""

    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/runners'
    data = {'runner_id' : runner_id}

    if not runner_is_enabled(project_path, runner_id, server, token):
        http_request(url, token, data)


def get_variables(project_path, server, token):
    """Returns a dict containing all environment variables set for the
    project.
    """
    pid      = lookup_project_id(project_path, server, token)
    url      = f'{server}/api/v4/projects/{pid}/variables'
    response = http_request(url, token)
    return {r['key'] : r['value'] for r in response}


def create_variable(project_path,
                    server,
                    token,
                    key,
                    value,
                    masked=False):
    """Creates a new variable on the project. """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/variables'
    data = dict(key=key, value=value, masked=masked)
    http_request(url, token, data=data)


def update_variable(project_path,
                    server,
                    token,
                    key,
                    value,
                    masked=False):
    """Updates the value of a variable on the project. """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/variables/{key}'
    data = {'value' : value, 'masked' : masked}
    http_request(url, token, data=data, method='PUT')


def create_or_update_variable(project_path,
                              server,
                              token,
                              key,
                              value,
                              **kwargs):
    """Create or update a variable on the project. """
    if key in get_variables(project_path, server, token):
        method = update_variable
    else:
        method = create_variable
    method(project_path, server, token, key, value, **kwargs)


def delete_variable(project_path, server, token, key):
    """Delete a variable from the project. """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/variables/{key}'
    http_request(url, token, method='DELETE')


def create_branch(project_path,
                  branch_name,
                  source_branch,
                  server,
                  token):
    """Create a new branch on the project, from the source branch/ref. """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/repository/branches'
    data = {
        'branch' : branch_name,
        'ref'    : source_branch,
    }
    http_request(url, token, data)


def open_merge_request(project_path,
                       source_branch,
                       message,
                       server,
                       token,
                       target_branch='master'):
    """Opens a merge request from the source brqanch to the target branch,
    on the given project.
    """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/merge_requests'
    data = {
        'id'                   : pid,
        'source_branch'        : source_branch,
        'target_branch'        : target_branch,
        'remove_source_branch' : 'true',
        'title'                : f'WIP: {source_branch}',
        'description'          : message,
    }
    http_request(url, token, data=data)


def trigger_pipeline(project_path,
                     ref,
                     server,
                     token,
                     variables=None):
    """Triggers a CI pipeline on the given project/ref. Returns the pipeline
    integer ID."""
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/pipeline'
    data = {'ref' : urlparse.quote_plus(ref)}

    if variables is not None:
        data['variables'] = [{'key' : k, 'value' : v}
                             for k, v in variables.items()]
    return http_request(url, token, data)


def get_pipeline_status(project_path, pipeline_id, server, token):
    """Returns information about the given pipeline. """
    pid = lookup_project_id(project_path, server, token)
    url = f'{server}/api/v4/projects/{pid}/pipelines/{pipeline_id}'
    return http_request(url, token)


def wait_on_pipeline(project_path, pipeline_id, server, token):
    """Waits until the given pipeline finishes, returning its final status. """

    start = time.time()

    while True:

        secs = int(time.time() - start)
        print(f'Waiting {secs} seconds...', flush=True)
        time.sleep(30)

        status = get_pipeline_status(
            project_path, pipeline_id, server, token)['status']

        if status not in ('created', 'waiting_for_resource',
                          'preparing', 'pending', 'running'):
            print(f'Pipeline has stopped - final status: {status}')
            break

        print(f'Pipeline status: {status}')


    return status


def get_project_jobs(project_path, server, token, scope=None, page=1):
    """Returns a list of pipeline jobs that have submitted for the given
    project.
    """
    pid = lookup_project_id(project_path, server, token)
    url = f'{server}/api/v4/projects/{pid}/jobs?page={page}'

    if scope is not None:
        url = f'{url}&scope={scope}'
    return http_request(url, token)


def trigger_job(project_path, job_id, server, token):
    """Start the specified CI job."""
    pid = lookup_project_id(project_path, server, token)
    url = f'{server}/api/v4/projects/{pid}/jobs/{job_id}/play'
    return http_request(url, token, method='POST')


def parse_gitlab_date(datestr):
    """2020-11-13T15:13:48.649Z
    """
    timestamp = datestr.split('.')[0]
    offset    = datestr.split('Z')[1]

    if offset == '': offset = 0
    else:            offset = float(offset)

    timestamp = datetime.datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S')
    offset    = datetime.timezone(datetime.timedelta(hours=offset))
    return timestamp.replace(tzinfo=offset)
