#!/usr/bin/env python
#
# Miscellaneous functions used by the Python-based FSL CI scripts.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import                    os
import                    re
import                    json
import                    shlex
import                    string
import                    tempfile
import                    datetime
import base64          as b64
import functools       as ft
import contextlib      as ctxlib
import subprocess      as sp
import urllib.parse    as urlparse
import urllib.request  as urlrequest

from typing import Dict

import jinja2          as j2
import                    yaml


@ctxlib.contextmanager
def tempdir():
    """Context manager to create, and change into, a temporary directory, and
    then afterwards delete it and change back to the original working
    directory.
    """
    with tempfile.TemporaryDirectory() as td:
        prevdir = os.getcwd()
        os.chdir(td)
        try:
            yield td
        finally:
            os.chdir(prevdir)


@ctxlib.contextmanager
def indir(dirname):
    """Context manager to change into a directory, and then afterwards
    change back to the original working directory.
    """
    prevdir = os.getcwd()
    os.chdir(dirname)
    try:
        yield
    finally:
        os.chdir(prevdir)


def sprun(cmd, **kwargs):
    """Runs the given command with subprocess.run. """
    print(f'Running {cmd}')
    cmd = shlex.split(cmd)
    return sp.run(cmd, check=True, **kwargs)


def http_request(url, token=None, data=None, method=None):
    """Submit a HTTP request to the given URL. """

    if method is None:
        if data is None: method = 'GET'
        else:            method = 'POST'

    print(f'{method} {url} ...')

    headers = {}

    if token is not None:
        headers['PRIVATE-TOKEN'] = token

    if data is not None:
        headers['Content-Type'] = 'application/json'
        data                    = json.dumps(data).encode('utf-8')

    request  = urlrequest.Request(
        url, headers=headers, data=data, method=method)
    response = urlrequest.urlopen(request).read()

    return json.loads(response)


def gen_recipe_path(project_path):
    """Generates a conda recipe project name from the given project.

    If the project name has one of the following forms (case doesn't matter):
      - "fsl-<project>"
      - "fsl_<project>"

    the recipe name will be "fsl-<project>". Otherwise, the recipe name will
    be "fsl-<project_name>".

    The full recipe path will be "fsl/conda/<recipe_name>"
    """

    # strip namespaced in case
    # fsl/<project> was passed in
    project_name = project_path.rsplit('/')[-1]
    match = re.match(r'^(fsl[_-])?(.*)$', project_name, flags=re.IGNORECASE)
    suffix = match.group(2).lower()

    return f'fsl/conda/fsl-{suffix}'


def get_recipe_url(project_name, server) -> Dict[str, str]:
    """Figures out the URL to the conda recipe associated with the specified
    FSL project.
    """


    # Figure out the name of the conda recipe associated with
    # this project. It defaults to "fsl-<project>", but can
    # be overridden via the FSLCONDA_RECIPE_URL environment
    # variable.
    url          = os.environ.get('FSLCONDA_RECIPE_URL')
    project_name = project_name.rsplit('/')[-1]

    if url is None:
        url = f'{server}/fsl/conda/fsl-{project_name}.git'

    path = urlparse.urlparse(url).path.replace('.git', '')[1:]
    name = path.rsplit('/', 1)[-1]

    return {'url' : url, 'name' : name, 'path' : path}


def is_standard_recipe_path(project_path, recipe_path):
    """Returns True if the given recipe path follows the "standard" convention
    of being "fsl/conda/fsl-<project_name>", False otherwise.
    """
    project_name = project_path.rsplit('/')[-1].lower()
    return recipe_path == f'fsl/conda/fsl-{project_name}'


@ft.lru_cache()
def lookup_project_id(project_path, server, token):
    """Look up the integer ID of a gitlab project from its fully qualified
    path.
    """
    project_path = urlparse.quote_plus(project_path)
    url          = f'{server}/api/v4/projects/{project_path}'
    return http_request(url, token)['id']


def lookup_namespace_id(namespace_path, server, token):
    """Look up the integer ID of a gitlab namespace from its fully qualified
    path.
    """
    url        = f'{server}/api/v4/namespaces'
    namespaces = http_request(url, token)

    for n in namespaces:
        if n['path'] == namespace_path:
            return n['id']

    raise ValueError(f'No namespace matching {namespace_path}')


def get_available_runners(project_path, server, token):
    """Returns a list of the IDs of all specific runners
    which are available to be used for the given project,
    and are not already enabled for it.
    """
    # get all specific runners available to the user
    url     = f'{server}/api/v4/runners'
    runners = http_request(url, token)

    # remove inactive ones, and ones that are already
    # enabled
    runners = [r for r in runners if r['online'] and r['active']]
    runners = [r for r in runners if not runner_is_enabled(
        project_path, r['id'], server, token)]

    return [r['id'] for r in runners]


def get_runner_metadata(runner_id, server, token):
    """Returns metadata for the specified runner. """
    url      = f'{server}/api/v4/runners/{runner_id}'
    response = http_request(url, token)
    return response


def get_runner_tags(runner_id, server, token):
    """Returns the list of tags for the specified runner. """
    return get_runner_metadata(runner_id, server, token)['tag_list']


def find_suitable_runners(project_path, tags, server, token):
    """Identifies runners with the specified set of tags which
    are available to be used on the given project.
    """
    def match(runner_tags):
        return all([t.lower() in runner_tags for t in tags])

    rids  = get_available_runners(project_path, server, token)
    rtags = [get_runner_tags(r, server, token) for r in rids]

    return [r for r, t in zip(rids, rtags) if match(t)]


def lookup_project_tags(project_path, server, token):
    """Return the a list of tags for the given project, or an empty list
    if the project has no tags.

    The tags are sorted such that the most recently updated tag is first
    in the list.
    """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/repository/tags'
    tags = http_request(url, token)
    tags = [t['name'] for t in tags]

    return tags


def is_acceptable(version):
    """Return True if the given version/tag is "acceptable" - it must
    be a sequence of integers, separated by periods, with an optional
    leading 'v'.
    """
    if version.lower().startswith('v'):
        version = version[1:]

    for part in version.split('.'):
        if not all([c in string.digits for c in part]):
            return False
    return True


def get_project_version(project_path, server, token):
    """Return the most recent version of the specified project, returning
    "YYMM.0" if the project does not have any tags.
    """
    tags    = lookup_project_tags(project_path, server, token)
    version = None

    if len(tags) >= 0 and is_acceptable(tags[0]):
        version = tags[0]

    if version is None:
        version = datetime.date.today().strftime('%y%m')
        version = f'{version}.0'

    return version


def list_project_branches(project_path, server, token):
    """Returns a list of all branches of the project. """
    pid      = lookup_project_id(project_path, server, token)
    url      = f'{server}/api/v4/projects/{pid}/repository/branches'
    response = http_request(url, token)
    return [r['name'] for r in response]


def get_project_metadata(project_path, server, token):
    """Returns metadata for the specifie project."""
    pid = lookup_project_id(project_path, server, token)
    url = f'{server}/api/v4/projects/{pid}/'
    return http_request(url, token)


def set_project_metadata(project_path, server, token, data):
    """Sets metadata for the specifie project."""
    pid = lookup_project_id(project_path, server, token)
    url = f'{server}/api/v4/projects/{pid}/'
    http_request(url, token, data, method='PUT')


@ft.lru_cache()
def download_file(
        project_path, filename, server, token, ref='master', text=True):
    """Download the specified from the specified branch/ref of the project. """
    pid      = lookup_project_id(project_path, server, token)
    filename = urlparse.quote_plus(filename)
    url      = f'{server}/api/v4/projects/{pid}/repository/files/'
    url      = f'{url}/{filename}?ref={ref}'
    contents = http_request(url, token)['content']

    # contents are base64 encoded
    contents = b64.b64decode(contents)

    # bytes or str
    if text:
        contents = contents.decode('utf-8')

    return contents


def update_file(project_path,
                filename,
                contents,
                message,
                server,
                token,
                branch='master'):
    """Update a file on the specified branch of the project. """

    pid      = lookup_project_id(project_path, server, token)
    filename =  urlparse.quote_plus(filename)
    url      = f'{server}/api/v4/projects/{pid}/repository/files/{filename}'
    data     = {
        'filename'       : urlparse.quote_plus(filename),
        'content'        : contents,
        'branch'         : branch,
        'commit_message' : message
    }
    http_request(url,  token, data=data, method='PUT')


def runner_is_enabled(project_path, runner_id, server, token):
    """Return True if the given runner is enabled for the given project,
    False otherwise.
    """
    pid     = lookup_project_id(project_path, server, token)
    url     = f'{server}/api/v4/projects/{pid}/runners'
    runners = http_request(url, token)
    rids    = [r['id'] for r in runners]

    return runner_id in rids


def enable_runner(project_path, runner_id, server, token):
    """Enables the specified runner on the specified project."""

    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/runners'
    data = {'runner_id' : runner_id}

    if not runner_is_enabled(project_path, runner_id, server, token):
        http_request(url, token, data)


def get_variables(project_path, server, token):
    """Returns a dict containing all environment variables set for the
    project.
    """
    pid      = lookup_project_id(project_path, server, token)
    url      = f'{server}/api/v4/projects/{pid}/variables'
    response = http_request(url, token)
    return {r['key'] : r['value'] for r in response}


def create_variable(project_path, server, token, key, value):
    """Creates a new variable on the project. """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/variables'
    data = dict(key=key, value=value)
    http_request(url, token, data=data)


def update_variable(project_path, server, token, key, value):
    """Updates the value of a variable on the project. """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/variables/{key}'
    data = {'value' : value}
    http_request(url, token, data=data, method='PUT')


def create_or_update_variable(project_path, server, token, key, value):
    """Create or update a variable on the project. """
    if key in get_variables(project_path, server, token):
        method = update_variable
    else:
        method = create_variable
    method(project_path, server, token, key, value)


def delete_variable(project_path, server, token, key):
    """Delete a variable from the project. """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/variables/{key}'
    http_request(url, token, method='DELETE')


def create_branch(project_path,
                  branch_name,
                  source_branch,
                  server,
                  token):
    """Create a new branch on the project, from the source branch/ref. """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/repository/branches'
    data = {
        'branch' : branch_name,
        'ref'    : source_branch,
    }
    http_request(url, token, data)


def open_merge_request(project_path,
                       source_branch,
                       message,
                       server,
                       token,
                       target_branch='master'):
    """Opens a merge request from the source brqanch to the target branch,
    on the given project.
    """
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/merge_requests'
    data = {
        'id'                   : pid,
        'source_branch'        : source_branch,
        'target_branch'        : target_branch,
        'remove_source_branch' : 'true',
        'title'                : f'WIP: {source_branch}',
        'description'          : message,
    }
    http_request(url, token, data=data)


def trigger_pipeline(project_path,
                     ref,
                     server,
                     token,
                     variables=None):
    """Triggers a CI pipeline on the given project/ref. Returns the pipeline
    integer ID."""
    pid  = lookup_project_id(project_path, server, token)
    url  = f'{server}/api/v4/projects/{pid}/pipeline'
    data = {'ref' : urlparse.quote_plus(ref)}

    if variables is not None:
        data['variables'] = [{'key' : k, 'value' : v}
                             for k, v in variables.items()]
    return http_request(url, token, data)['id']


def get_pipeline_status(project_path,
                        pipeline_id,
                        server,
                        token):
    pid      = lookup_project_id(project_path, server, token)
    url      = f'{server}/api/v4/projects/{pid}/pipelines/{pipeline_id}'
    response = http_request(url, token)

    return response


def load_meta_yaml(filename):
    """Load a conda recipe meta.yaml file."""

    env = {
        'os'                 : os,
        'load_setup_py_data' : lambda : {},
    }

    with open(filename, 'rt') as f:
        template = j2.Template(f.read())

    meta = template.render(**env)
    meta = yaml.load(meta, Loader=yaml.Loader)

    return meta
