#!/usr/bin/env python
#
# project_dependencies.py - Identify dependencies of a FSL project.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import               os
import os.path    as op
import subprocess as sp
import itertools  as it
import               re
import               sys
import               shlex
import               shutil

from fsl_ci_utils import (tempdir, indir)


BINARY_DEPENDENCY_REPLACEMENTS = {
    'boost'                 : 'boost-cpp',
    'libprob'               : 'cprob',
    'ciftiio'               : 'CiftiLib',
    'CiftiLib-master'       : 'CiftiLib',
    'CiftiLib-master/Nifti' : 'CiftiLib',
    'flirt/flirtsch'        : 'flirt',
    'miscvis/luts+pics'     : 'miscvis',
    'libxml++-2.6'          : 'libxmlpp',
}
"""Used by get_binary_dependencies to replace some common false positives.
"""


EXCLUDE_FROM_BINARY_DEPOENDENCIES = [
    'x86_64-linux-gnu',
    'linux'
]
"""Any paths which are listed as binary (C/C++) dependencies of a project,
and which contain any of the items in this list as sub-strings, are excluded.
See get_binary_dependencies.
"""


# These identifiers are returned by the filetype function.
SHELL   = 'shell'
PYTHON  = 'python'
BINARY  = 'binary'
DATA    = 'data'
UNKNOWN = 'unknown'


def sprun(cmd):
    cmd = shlex.split(cmd)
    out = sp.run(cmd, check=True, capture_output=True).stdout
    return out.decode().strip()


def which(command):
    """Run "which command". """
    return sprun(f'which {command}')


def filetype(filename):
    """Attempt to identify the type of the file, either a shell script,
    compiled executable, python script, or unknown.
    """

    filename = op.expandvars(filename)
    ftype    = sprun(f'file -L "{filename}"').lower()

    if   'shell'  in ftype: return SHELL
    elif 'python' in ftype: return PYTHON
    elif 'x86-64' in ftype: return BINARY
    else:                   return UNKNOWN


def find_source_project(cmdname):
    """Try and identify the FSL project which provides the given command.
    """
    srcdir  = op.join(os.environ['FSLDIR'], 'src')
    results = sprun(f'find {srcdir} -name {cmdname} -or -name {cmdname}.cc')

    results = [r.strip()             for r in results.split()]
    results = [op.relpath(r, srcdir) for r in results]
    results = [op.dirname(r)         for r in results]

    results = sorted(set(results))
    results = [r for r in results if r != '']

    return results


def get_project_executables(project_dir):
    """Return a list of executables provided by the given FSL project."""

    makefile = op.join(project_dir, 'Makefile')
    if not op.exists(makefile):
        return []

    lines  = []
    concat = False
    with open(makefile) as f:
        for line in f:
            line = line.strip()
            if not concat:
                lines.append(line)
            else:
                lines[-1] = lines[-1][:-2] + ' ' + line
            concat = line.endswith('\\')

    patterns = [
        r'XFILES *=',
        r'SCRIPTS *='
    ]

    xlines = [l for l in lines if any([re.match(p, l) for p in patterns])]
    xfiles = list(it.chain(*[l.split('=')[1].split() for l in xlines]))
    xfiles = [x.strip() for x in xfiles]
    return xfiles


def _get_shell_dependencies(filename):
    """Return the FSL dependencies of the given shell script. Used by
    get_shell_dependencies.
    """

    filename = op.expandvars(filename)
    ftype    = filetype(filename)

    if ftype != SHELL:
        return []

    with open(filename, 'rt', encoding='UTF-8') as f:
        lines = f.readlines()
        lines = [l.strip() for l in lines]

    pats = [
        r'(\$FSLDIR/[^ \'"`\)]+)',
        r'(\$\{FSLDIR\}/[^ \'"`\)]+)']
    pats = [re.compile(p) for p in pats]

    deps = []
    for line in lines:
        for pat in pats:
            match = re.search(pat, line)
            if match:
                dep = match.groups(1)[0]
                if op.expandvars(dep) != filename:
                    deps.append(dep)
                break
    return list(sorted(set(deps)))


def get_shell_dependencies(filename):
    """Recursively identify the FSL dependencies of the given shell script.
    """

    cmddeps = _get_shell_dependencies(filename)
    subdeps = []

    for dep in cmddeps:
        subdeps.extend(get_shell_dependencies(dep))

    return cmddeps + subdeps


def get_binary_dependencies(project_dir):
    """Attempt to identify binary (C/C++) dependencies of the given
    FSL project.
    """

    with tempdir():
        shutil.copytree(project_dir, 'project')
        with indir('project'):
            sprun('rm -f depend.mk')
            sprun('make depend.mk')

            with open('depend.mk') as f:
                lines = f.readlines()

    lines    = [l.strip(' \\\n') for l in lines]
    lines    = [l for l in lines if '/include/' in l]
    lines    = [l.split('/include/')[1] for l in lines]
    projects = [l.split('/')[0].strip() for l in lines]
    projects = [p for p in projects if not p.endswith('.h')]
    projects = [p for p in projects
                if p not in EXCLUDE_FROM_BINARY_DEPOENDENCIES]
    projects = [BINARY_DEPENDENCY_REPLACEMENTS.get(p, p) for p in projects]

    return sorted(set(projects))


def get_project_dependencies(project_dir):
    """Attempt to identify all of the dependencies of the given FSL project.
    """

    exes = get_project_executables(project_dir)
    deps = get_binary_dependencies(project_dir)

    for exe in exes:
        if filetype(exe) == SHELL:
            edeps = get_shell_dependencies(exe)
            edeps = [find_source_project(e) for e in edeps]
            deps.extend(edeps)

    return sorted(set(deps))


def main():
    if ('FSLDIR' not in os.environ) or ('FSLDEVDIR' not in os.environ):
        print('FSLDIR and FSLDEVDIR must be set!')
        sys.exit(1)

    if len(sys.argv) != 2:
        print('Usage: project_dependencies.py project_dir')
        sys.exit(1)

    project_dir = sys.argv[1]
    deps        = get_project_dependencies(project_dir)

    for d in deps:
        print(d)



if __name__ == '__main__':
    main()
