#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import inspect


# IMPLEMENTATION

def create_error_message(exc, msg, default_msg, *args):
    try:
        caller = inspect.stack()[2][3]
        if msg is not None:
            # args = ("(caller: {}): ".format(caller) + msg,) + args
            args = (msg,) + args
        else:
            # args = ("(caller: {}): ".format(caller) + default_msg,) + args
            args = (default_msg,) + args
    except Exception:
        args = (default_msg,) + args
        raise
    finally:
        super(type(exc), exc).__init__(*args)


class ConstructionError(Exception):

    def __init__(self, msg, *args):
        try:
            caller = inspect.stack()[1][0].f_locals["self"].__class__.__name__
            args = ("{0}: {1}".format(caller, msg),) + args
        except:
            pass
        finally:
            super().__init__(*args)


class ArgumentError(Exception):
    """ Exception raised during the validation process, because the argument
    is required by the caller but it is invalid. """

    def __init__(self, msg=None, *args):
        default_msg = "Argument error."
        create_error_message(self, msg, default_msg, *args)


class CacheError(Exception):
    """ Exception that is raised upon unexpected result of Cache operation. """

    def __init__(self, msg=None, *args):
        default_msg = "Unexpected error while using object cache."
        create_error_message(self, msg, default_msg, *args)


class DomainError(Exception):
    """ Exception raised when an error within a Domain object would interfere
    with further execution of the program. """

    def __init__(self, msg=None, *args):
        default_msg = "Domain error."
        create_error_message(self, msg, default_msg, *args)


class TransformationError(Exception):
    """ Exception raised when an error within a Transformation object would
    interfere with further execution of the program. """

    def __init__(self, msg=None, *args):
        default_msg = "Transformation error."
        create_error_message(self, msg, default_msg, *args)


class OptionError(Exception):
    """ Exception raised when invalid or missing option in the settings file
    prevents further execution of the program. """

    def __init__(self, msg=None, *args):
        default_msg = "Option error."
        create_error_message(self, msg, default_msg, *args)


class CostError(Exception):
    """ Exception raised when invalid or missing option in the settings file
    prevents further execution of the program. """

    def __init__(self, msg=None, *args):
        default_msg = "Cost function error."
        create_error_message(self, msg, default_msg, *args)


class OptimisationError(Exception):
    """ Exception raised when invalid or missing option in the settings file
    prevents further execution of the program. """

    def __init__(self, msg=None, *args):
        default_msg = "Optimisation error."
        create_error_message(self, msg, default_msg, *args)


class BroadcastError(Exception):
    """ Exception raised when invalid or missing option in the settings file
    prevents further execution of the program. """

    def __init__(self, msg=None, *args):
        default_msg = "Broadcast error."
        create_error_message(self, msg, default_msg, *args)


if __name__ == "__main__":  # pragma: no cover
    print("""This module is not intended for execution.""")
