#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import numpy as np


# TIRL IMPORT

from tirl.interpolators.interpolator import Interpolator
from tirl.cmodules import fslinterpolator


# IMPLEMENTATION

class FSLInterpolator(Interpolator):
    """
    This class implements trilinear interpolation with constant (=0)
    extrapolation. The class demonstrates how custom interpolators can be
    used with TIRL.

    """
    RESERVED_KWARGS = ("source", "tensor_axes", "threads", "verbose",
                       "ipkwargs", "kwargs", "radius", "hold")

    def __init__(self, source, tensor_axes=(), threads=-1, verbose=False,
                 ipkwargs=None, **kwargs):
        """
        FSLInterpolator initialiser.

        """
        # Invoke the initialiser of the superclass
        kwargs = {k: v for k, v in kwargs.items()
                  if k not in self.RESERVED_KWARGS}
        super(FSLInterpolator, self).__init__(
            source=source, tensor_axes=tensor_axes, radius=1, hold=False,
            threads=threads, verbose=verbose, ipkwargs=ipkwargs, **kwargs)

        # Set class-specific properties
        # self.radius = 1

    def interpolate(self, coordinates, input_array=None, **kwargs):
        """
        Implementation of the interpolation algorithm.

        :param coordinates:
            (n_points, n_dimensions) table of coordinates where the input array
            should be evaluated by interpolation
        :type coordinates: np.ndarray
        :param input_array:
            Array of known values. If the source data is bigger than memory,
            input_array will correspond to a chunk of the original array.
            Chunks are created with overlap and the values that are
            interpolated at the chunk boundary will be automatically discarded.
            The extent of the overlap is defined by
        :type input_array: np.ndarray
        :param kwargs:

        :returns: (n_points,) interpolated values
        :rtype: np.ndarray

        """
        if isinstance(coordinates, (tuple, list)):
            coordinates = np.asarray(coordinates).reshape((1, -1))
        try:
            res = fslinterpolator(input_array, coordinates)
        except:
            print("Interpolation failed with the following input: ",
                  input_array.shape, coordinates.shape, kwargs)
            raise
        else:
            return res
