#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# TIRL IMPORTS

from tirl.regularisers.regulariser import Regulariser
from tirl.optimisation.optgroup import OptimisationGroup
from tirl.transformations.transformation import Transformation


# IMPLEMENTATION

class TransformationRegulariser(Regulariser):
    """
    TransformationRegulariser - a class for regularisation terms that are
    computed solely from transformation parameters.

    """
    RESERVED_KWARGS = ("transformation", "weight", "logger", "metaparameters")

    def __init__(self, transformation, weight=1.0, logger=None,
                 **metaparameters):
        """
        Initialisation of TransformationRegulariser.

        :param transformation:
            Transformation object whose parameters are used to compute to the
            regularisation term.
        :type transformation: Transformation or OptimisationGroup
        :param weight: Regularisation weight.
        :type weight: Real
        :param logger:
            If None (default), logs will be redirected to the RootLogger
            instance. You may alternatively specify a different Logger instance
            or its name.
        :type logger: None or str or Logger
        :param kwargs:
            Additional keyword arguments required to compute the
            regularisation term.
        :type kwargs: Any

        """
        # Check whether the input has suitable parameters
        if not isinstance(transformation, (Transformation, OptimisationGroup)):
            raise TypeError(
                f"Expected Transformation or OptimisationGroup "
                f"for {self.__class__.__name__}, "
                f"got {transformation.__class__.__name__} instead.")

        # Call superclass initialisation
        super(TransformationRegulariser, self).__init__(
            transformation.parameters, weight=weight, logger=logger,
            **metaparameters)

        # Obtain a hook to the transformation or optimisation group, so that
        # they can be identified by the optimiser.
        self.transformation = transformation

    @property
    def transformation(self):
        return self.metaparameters.get("transformation")

    @transformation.setter
    def transformation(self, tx):
        if isinstance(tx, Transformation):
            self.metaparameters.update(transformation=tx)
        else:
            raise TypeError(f"Expected a Transformation (or subclass), "
                            f"got {tx} instead.")


if __name__ == "__main__":
    print("""This module is not intended for execution.""")
