#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar
# Date: 9 June 2020


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


from tirl.scripts.mnd import __version__
__tirlscript__ = True


# DEPENDENCIES

import os
import sys
import argparse
import numpy as np
from tirl.timage import TImage
from skimage.color import rgb2hsv
from skimage.filters import threshold_otsu
from skimage.measure import label, regionprops


# DEFINITIONS

OPTIONS = {
    "distance": 0.25,
    "channel": 0
}


# IMPLEMENTATION

def crop_centre(args):
    """ Main program code. """

    img = TImage(args.image).tensors[:3]
    img.resample(args.scale, copy=False)
    hue = rgb2hsv(img.data)[..., OPTIONS["channel"]]
    binary = np.where(hue < threshold_otsu(hue), 255, 0)
    labelimg = label(binary, background=0)
    regions = regionprops(labelimg, hue)
    mask = np.zeros_like(binary)

    # Select pixels that belong to the ROI
    for region in regions:
        if region.label == 0:
            continue
        radius = np.mean(np.multiply(OPTIONS["distance"], img.vshape))
        centre = np.divide(img.vshape, 2)
        dist = np.linalg.norm(np.subtract(centre, region.centroid))
        if dist < radius:
            mask[labelimg == region.label] = 1

    # Get bounding box from binary mask
    mask[mask > 0.5] = 1
    mask[mask <= 0.5] = 0

    # TImage.fromarray(mask).preview()

    # Determine minimum bounding box
    minrow = np.inf
    mincol = np.inf
    maxrow = -np.inf
    maxcol = -np.inf
    for region in regionprops(label(mask, background=0)):
        minr, minc, maxr, maxc = region.bbox
        minrow = int(min(minrow, minr))
        mincol = int(min(mincol, minc))
        maxrow = int(max(maxrow, maxr))
        maxcol = int(max(maxcol, maxc))

    # Crop with margin
    img.resample(1, copy=False)
    h, w = img.vshape
    minrow /= args.scale
    mincol /= args.scale
    maxrow /= args.scale
    maxcol /= args.scale
    height = maxrow - minrow
    width = maxcol - mincol
    left, right, top, bottom = args.margin
    left = int(width * left)
    right = int(width * right)
    top = int(height * top)
    bottom = int(height * bottom)
    minrow = int(max(minrow - top, 0))
    maxrow = int(min(maxrow + bottom, h))
    mincol = int(max(mincol - left, 0))
    maxcol = int(min(maxcol + right, w))
    imdata = img.data
    img = TImage.fromarray(
        imdata[minrow:maxrow, mincol:maxcol], tensor_axes=(2,))

    # Generate file name
    fname, ext = os.path.splitext(args.image)
    fname = fname.rstrip(".") + "_crop" + f".{ext}"
    
    # Flip image if needed
    if args.flip:
        fname = fname.replace("_crop", "_crop_flip")
        img.data[...] = img.data[:, ::-1]

    # Save file
    img.snapshot(fname, overwrite=True)


def create_cli():
    """
    Creates command-line interface.

    """
    descr = "Segments brain slice from dissection photograph."
    parser = argparse.ArgumentParser(prog="crop_centre", description=descr)

    parser.add_argument("-i", "--image", metavar="file", type=str, default="",
                        required=True, help="Input image.")
    parser.add_argument("-s", "--scale", metavar="factor", type=float,
                        default=0.25, required=False, help="Scaling factor.")
    parser.add_argument("-m", "--margin", type=float, nargs=4,
                        default=[0.1, 0.1, 0.1, 0.1], required=False,
                        help="Relative margins (left, right, top, bottom).")
    parser.add_argument("--flip", action="store_true", default=False,
                        required=False, help="Flip input horizontally.")

    return parser


def main(*args):
    """ Main program code. """

    parser = create_cli()
    if args:
        crop_centre(parser.parse_args(args))
    else:
        parser.print_help()


if __name__ == "__main__":
    main(*sys.argv[1:])
