#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar
# Date: 21 June 2020


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


from tirl.scripts.mnd import __version__
__tirlscript__ = False


# DEPENDENCIES

import logging
import numpy as np
from attrdict import AttrMap


# DEFINITIONS

logger = logging.getLogger("tirl.scripts.mnd.image")  # do not change this

EPS = np.finfo(np.float32).eps

DEFAULT_MASK_MODULE = {
    "file": None,
    "normalise": True,
    "function": None,
    "automask": {
        "thr": 0,
        "uthr": 1.0
    }
}


# IMPLEMENTATION

def set_mask(img, scope=None, **maskparammodule):
    """
    Sets TImage mask based on the inputs in the corresponding mask parameter
    module.

    """
    img.mask = get_mask(img, scope=scope, **maskparammodule)


def get_mask(img, scope=None, **maskparammodule):
    """
    Creates a TImage mask based on the inputs in the corresponding mask
    parameter module.

    """
    # Get mask parameters from module (user input overrides defaults)
    q = DEFAULT_MASK_MODULE.copy()
    q.update(maskparammodule)
    # Set scope (where function definitions will be looked in)
    scope = globals() if scope is None else scope
    assert isinstance(scope, dict), "Scope must be a dict."
    q = AttrMap(q)

    # Load from image file and normalise if asked
    if q.file is not None:
        from tirl.tfield import TField
        mask = TField.fromfile(q.file, dtype="<f4")
        mask.reduce_tensors()
        if q.normalise:
            mask = mask / np.max(mask.data)
        return mask.data

    # Create mask by calling a function in 'scope'
    elif q.function is not None:
        from pydoc import locate
        func = scope.get(q.function, locate(q.function))
        mask = func(img)
        return mask

    # Create a mask by thresholding the normalised intensities
    elif (q.automask.thr != 0.0) or (q.automask.uthr != 1.0):
        mask = mask_from_threshold(
            img, thr=q.automask.thr, uthr=q.automask.uthr)
        return mask

    else:
        return img.mask


def mask_from_threshold(img, thr=None, uthr=None):
    """
    Automatic masking based on foreground segmentation by thresholding.
    Normalised intensities that fall between the specified lower and upper
    intensity thresholds are segmented as foreground, and the rest of the
    pixels are masked out.

    """
    # Winsorize image data (handle outliers)
    tmp = img.copy()
    tmp.reduce_tensors()
    low = np.percentile(tmp.data, 0.5)
    high = np.percentile(tmp.data, 99.5)
    tmp.data[...] = np.clip(tmp.data, low, high)
    amin, amax = np.min(tmp.data), np.max(tmp.data)

    # Generate mask by thresholding
    mask = np.ones(tmp.vshape, dtype=np.float32)
    if thr is not None:
        assert 1 + EPS >= thr >= 0 - EPS
        mask[((tmp.data - amin) / (amax - amin)) < thr] = 0
    if uthr is not None:
        assert 1 + EPS >= thr >= 0 - EPS
        mask[((tmp.data - amin) / (amax - amin)) > uthr] = 0

    return mask


def perform_image_operations(img, *actions, scope=None, other=None, **kwargs):
    """
    Calls the sequence of action functions on the specified TImage.

    The input image should not be modified by the action functions. The result
    of each action is used as the input for the next action. The action
    functions should have the following signature:

        actionfunc(img: TImage, other=None, history=()) -> result: TImage

    The result of the last action is returned by this function.

    :param img: TImage on which the actions are performed
    :type img: TImage
    :param actions:
        Sequence of functions from the TIRL namespace defined by their full
        module path. For example, to call the 'filter' function from the user
        module 'histology' in package 'pkg', enter
        'tirl.usermodules.pkg.histology.filter'.
    :type actions: str
    :param scope: where functions are looked for
    :type scope: dict
    :param other: other TImage that is used in the registration (optional)
    :type other: TImage
    :param kwargs:
        Further arguments that are necessary for the image operation.
        The global configurations (as loaded from a configuration file) should
        be added under the "cnf" key by the caller.
    :type kwargs: Any

    :returns: pre-processed image
    :rtype: TImage

    """
    if not actions:
        return img

    from pydoc import locate

    action_history = []
    lastimg = img
    for action in actions:
        func = scope.get(action, locate(action))
        res = func(lastimg, other=other, history=action_history, **kwargs)
        lastimg = res or lastimg
        action_history.append((action, lastimg))
    else:
        from tirl.timage import TImage
        assert isinstance(lastimg, TImage), \
            "Actions must not change the TImage type"
        return lastimg


def flip_x(img, **kwargs):
    """
    Flips the image along the horizontal axis.

    """
    from tirl.transformations.linear.affine import TxAffine
    pos = img.domain.map_voxel_coordinates([np.subtract(img.vshape, 1)])[0]
    swapmat = np.eye(img.vdim + 1)
    swapmat[0, 0] *= -1
    swapmat[0, -1] = pos[0]
    swap = TxAffine(swapmat[:-1, :], name="flip_x")
    img.domain.chain.prepend(swap)
    # img.data[...] = img.data[::-1]
    # if img.tmask() is not None:
    #     img.mask[...] = img.mask[::-1]


def flip_y(img, **kwargs):
    """
    Flips the image along the horizontal axis.

    """
    from tirl.transformations.linear.affine import TxAffine
    pos = img.domain.map_voxel_coordinates([np.subtract(img.vshape, 1)])[0]
    swapmat = np.eye(img.vdim + 1)
    swapmat[1, 1] *= -1
    swapmat[1, -1] = pos[1]
    swap = TxAffine(swapmat[:-1, :], name="flip_y")
    img.domain.chain.prepend(swap)
    # img.data[...] = img.data[:, ::-1, ...]
    # if img.tmask() is not None:
    #     img.mask[...] = img.mask[:, ::-1]


def rgb2yiq(timg):
    """
    Convert RGB to YIQ, or leave as grayscale.

    """
    from tirl.operations.tensor import TensorOperator

    if timg.tsize >= 3:
        mat = np.asarray([[ 0.299,   0.587,   0.114],
                          [0.5959, -0.2746, -0.3213],
                          [0.2115, -0.5227,  0.3112]])
        return TensorOperator(lambda t: np.einsum("ij,...j", mat, t))(timg)
    else:
        return timg
