#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar
# Date: 23 June 2020


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


__version__ = "2020.7.1"
__tirlscript__ = True


# DEPENDENCIES

import os
import sys
import argparse


# TIRL IMPORTS

import tirl


# IMPLEMENTATION

def sliceview(args):
    """
    Main program code.

    """
    from mayavi import mlab
    from scipy.spatial import Delaunay

    images = []
    for imfile in args.slice:
        if os.path.isfile(imfile):
            images.append(tirl.load(imfile))
        else:
            raise FileNotFoundError(f"Image file not found: {imfile}")

    fig = mlab.figure(1, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5))

    # Volume
    if os.path.isfile(args.volume):
        mri = tirl.load(args.volume)
        points = mri.domain.get_voxel_coordinates()
        x, y, z = points.T
        x = x.reshape(mri.data.shape)
        y = y.reshape(mri.data.shape)
        z = z.reshape(mri.data.shape)
        mlab.volume_slice(x, y, z, mri.data, colormap="gray")
    else:
        mri = None

    # Slices
    for img in images:
        img = img.tensors[0]
        tri = Delaunay(img.domain.get_voxel_coordinates())
        points = img.domain.get_physical_coordinates()
        if mri is not None:
            points = mri.domain.map_physical_coordinates(points)
        tmesh = mlab.triangular_mesh(*points.T, tri.simplices, colormap="gray")
        tmesh.mlab_source.scalars = img.data.reshape(-1)
        tmesh.actor.mapper.scalar_visibility = True

    mlab.show()


def create_cli():
    """
    Creates command-line interface.

    """
    usage = """./sliceview --slice <slice> [slices ...] --volume <volume>"""
    descr = "View registered slices in a 3D window with the volume."

    parser = argparse.ArgumentParser(
        prog="sliceview", usage=usage, description=descr)
    parser.add_argument("--slice", metavar="timg", help="Registered 2D TImage.",
                        nargs="+", required=False, type=str)
    parser.add_argument("--volume", metavar="timg", help="3D TImage.",
                        required=False, type=str, default="")

    return parser


def main(*args):
    """ Main program code. """

    parser = create_cli()
    if args:
        sliceview(parser.parse_args(args))
    else:
        parser.print_help()


if __name__ == "__main__":
    main(*sys.argv[1:])
