#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   _______   _____   _____    _
#  |__   __| |_   _| |  __ \  | |
#     | |      | |   | |__) | | |
#     | |      | |   |  _  /  | |
#     | |     _| |_  | | \ \  | |____
#     |_|    |_____| |_|  \_\ |______|
#
# Copyright (C) 2018-2020 University of Oxford
# Part of the FMRIB Software Library (FSL)
# Author: Istvan N. Huszar


# FMRIB Software Library, Release 6.0.4 (c) 2020, The University of Oxford
# (the "Software")
# 
# The Software remains the property of the University of Oxford ("the
# University").
# 
# The Software is distributed "AS IS" under this Licence solely for
# non-commercial use in the hope that it will be useful, but in order
# that the University as a charitable foundation protects its assets for
# the benefit of its educational and research purposes, the University
# makes clear that no condition is made or to be implied, nor is any
# warranty given or to be implied, as to the accuracy of the Software,
# or that it will be suitable for any particular purpose or for use
# under any specific conditions. Furthermore, the University disclaims
# all responsibility for the use which is made of the Software. It
# further disclaims any liability for the outcomes arising from using
# the Software.
# 
# The Licensee agrees to indemnify the University and hold the
# University harmless from and against any and all claims, damages and
# liabilities asserted by third parties (including claims for
# negligence) which arise directly or indirectly from the use of the
# Software or the sale of any products based on the Software.
# 
# No part of the Software may be reproduced, modified, transmitted or
# transferred in any form or by any means, electronic or mechanical,
# without the express permission of the University. The permission of
# the University is not required if the said reproduction, modification,
# transmission or transference is done without financial return, the
# conditions of this Licence are imposed upon the receiver of the
# product, and all original and amended source code is included in any
# transmitted product. You may be held legally responsible for any
# copyright infringement that is caused or encouraged by your failure to
# abide by these terms and conditions.
# 
# You are not permitted under this Licence to use this Software
# commercially. Use for which any financial return is received shall be
# defined as commercial use, and includes (1) integration of all or part
# of the source code or the Software into a product for sale or license
# by or on behalf of Licensee to third parties or (2) use of the
# Software or any derivative of it for research with the final aim of
# developing software products for sale or license to a third party or
# (3) use of the Software or any derivative of it for research with the
# final aim of developing non-software products for sale or license to a
# third party, or (4) use of the Software to provide any service to an
# external organisation for which payment is received. If you are
# interested in using the Software commercially, please contact Oxford
# University Innovation ("OUI"), the technology transfer company of the
# University, to negotiate a licence. Contact details are:
# Innovation@innovation.ox.ac.uk quoting reference BS/9564.
# 


# DEPENDENCIES

import numpy as np
from numbers import Integral
from scipy.interpolate import Rbf


# TIRL IMPORTS

import tirl.utils as tu
import tirl.settings as ts
from tirl.domain import Domain
from tirl.cmodules.finv import local_affine
from tirl.transformations.nonlinear.displacement import TxDisplacementField


# DEFINITIONS

from tirl.constants import *


# IMPLEMENTATION

class TxRbfDisplacementField(TxDisplacementField):
    """
    TxRbfDisplacementField class -
        a deformation field defined on a sparse domain.

    """
    RESERVED_KWARGS = ("field", "domain", "interpolator", "bounds", "lock",
                       "name", "invertible", "mode", "vectorder", "dense_shape",
                       "metaparameters")

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ INITIALISATION ~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

    def __init__(self, field, domain=None, interpolator=None, bounds=None,
                 lock=None, name=None, invertible=True, mode=NL_ABS,
                 vectorder=None, dense_shape=None, **metaparameters):
        """
        Initialisation of TxRbfDisplacementField.

        To qualify, this class must be initialised with a Domain that is
        non-compact, or a TField that is defined on such a Domain.

        :param field:
            Vector field defined as a TField or an NDArray. If the input is a
            TField, the initialiser will enforce TENSOR_MAJOR layout on the
            input field. The input field must be defined on a non-compact
            domain. If the input is an NDArray, the shape must be (v, ...),
            where v is the dimensionality of the displacement vectors. By
            default, v should be equal to the number of spatial dimensions of
            the transformation domain. If the vectors are only defined on a
            subset of the spatial dimensions, or the orientation of the vector
            space is different from that of the transformation domain, this
            must be indicated by setting the "vectorder" argument accordingly
            (see there).
        :type field: TField or np.ndarray
        :param domain:
            Domain that defines the extent of the transformation. This
            argument is ignored if 'field' is a TField. If the field is an
            array, the domain must be specified by a non-compact Domain.
        :type domain: None or Domain
        :param interpolator:
            Field interpolator instance. This argument takes precedence over
            the TField specification. If None (default), and the field input is
            an array, the DEFAULT_NONCOMPACT_INTERPOLATOR will be used.
        :type interpolator: Interpolator or None
        :param bounds:
            Bounds specification for the field parameters.
        :type bounds: tuple
        :param metaparameters: metaparameters of the transformation
        :type metaparameters: Any

        """
        d = domain or getattr(field, "domain", None)
        if not isinstance(d, Domain) or d.iscompact:
            raise TypeError(f"{self.__class__.__name__} must be initialised "
                            f"with a non-compact domain.")

        super(TxRbfDisplacementField, self).__init__(
            field, domain=domain, interpolator=interpolator, bounds=bounds,
            lock=lock, name=name, invertible=invertible, mode=mode,
            vectorder=vectorder, **metaparameters)

        # Set class-specific metaparameters
        self.dense_shape = dense_shape

    # Make sure that this class switches to the dense representation in all
    # arithmetics with other transformations.

    def __add__(self, other):
        return self.dense().__add__(other)

    def __radd__(self, other):
        return self.dense().__radd__(other)

    def __sub__(self, other):
        return self.dense().__sub__(other)

    def __mul__(self, other):
        return self.dense().__mul__(other)

    def __rmul__(self, other):
        return self.dense().__rmul__(other)

    @property
    def dense_shape(self):
        return self.metaparameters.get("dense_shape")

    @dense_shape.setter
    def dense_shape(self, shape):
        # Set default dense shape if the input is None
        # This assumes proper voxel coordinates, not floats in [0, 1].
        if shape is None:
            shape = np.ceil(np.max(
                self.domain.get_voxel_coordinates(), axis=0)).astype(np.int)
            if np.any(shape < 1):
                from warnings import warn
                warn(f"Dense domain shape for {self.name} is too small. Make "
                     f"sure that voxel coordinates are true voxel coordinates, "
                     f"not normalised floating-point values.")

        if hasattr(shape, "__iter__") and \
                all(isinstance(ix, Integral) for ix in shape):
            shape = tuple(shape)
            if len(shape) == self.domain.ndim:
                self.metaparameters.update(dense_shape=shape)
            else:
                raise AssertionError(
                    "The shape of the dense Domain must have the "
                    "same number of dimensions as the sparse "
                    "domain.")
        else:
            raise ValueError(f"Expected a sequence of ints for dense_shape, "
                             f"got {shape} instead.")


    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PUBLIC METHODS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

    def copy(self):
        field = self.field().copy()
        metaparameters = tu.rcopy(self.metaparameters)
        meta = {k: v for k, v in metaparameters.items()
                if k not in self.RESERVED_KWARGS}
        name = self.name
        bounds = (self.parameters.lower_bounds, self.parameters.upper_bounds)
        tx = type(self)(field, bounds=bounds, lock=self.parameters.locked,
                        name=name, invertible=self.invertible, mode=self.mode,
                        vectorder=self.vectorder, dense_shape=self.dense_shape,
                        **meta)
        return tx

    def dense(self):
        """
        Returns a dense respresentation of the deformation field.

        """
        domain = Domain(self.dense_shape, *self.domain.chain,
                        offset=self.domain.offset[1:], name="dense")
        field = self.std().absolute().field(domain)
        return TxDisplacementField(field)

    @classmethod
    def _load(cls, dump):

        params = dump.get("parameters")
        meta = dump.get("metaparameters")

        # Create TField
        domain = meta.pop("domain")
        vectorder = meta.pop("vectorder")
        interpolator = meta.pop("interpolator")
        interpolator.values = params.parameters.reshape((-1, *domain.shape))
        assert interpolator.values.base is params.parameters.base
        name = meta.pop("name")
        invertible = meta.pop("invertible")
        mode = meta.pop("mode")
        dense_shape = meta.pop("dense_shape", None)

        from tirl.tfield import TField

        field = TField(
            domain, tensor_shape=(len(vectorder),), order=TENSOR_MAJOR,
            dtype=ts.DEFAULT_FLOAT_TYPE, buffer=params.parameters, offset=0,
            interpolator=interpolator, name=None, storage=MEM)

        # Create transformation object
        bounds = params.get_bounds()
        lock = params.locked
        meta = {k: v for k, v in meta.items() if k not in cls.RESERVED_KWARGS}
        obj = cls(field, bounds=tuple(bounds.T), lock=lock, name=name,
                  invertible=invertible, mode=mode, vectorder=vectorder,
                  dense_shape=dense_shape, **meta)

        return obj

    def matrix(self, coords=None):
        """
        Returns the local affine approximation to the transformation at the
        specified physical locations.

        """
        # Obtain absolute deformation field on an equivalent dense domain
        name = self.domain.name + "_dense"
        dense_domain = Domain(self.dense_shape, *self.domain.chain,
                              offset=self.domain.offset[1:], name=name)
        if self.mode == NL_ABS:
            field = self.field(dense_domain)
        else:
            field = self.rel2abs(self.field(dense_domain))

        # Locations where the local transformation is defined exactly
        gridlocations = field.domain.get_physical_coordinates()
        # spacedim = gridlocations.shape[-1]

        # Obtain the locations within the field
        # where the matrix should be evaluated.
        if coords is None:
            coords = gridlocations
        else:
            coords = np.atleast_2d(coords)

        # Choose local affine routine (C/Cython functions) depending on the
        # dimensions of the physical space, in which the current
        # transformation operates. Note that this may differ from vectdim: the
        # transformation may only operate on x and y coordinates of 3D space.
        affines = local_affine(
            field=field.data, newlocations=coords, gridlocations=gridlocations,
            output=None, vectorder=self.vectorder, maxiter=1000)

        # Fix rank deficiency: if the transformation does not update all
        # coordinates of the input points, the estimated local affine
        # transformation will be rank-deficient, with a zero determinant. This
        # is fixed here by replacing the 0s on the diagonal with 1s.
        # if self.vectdim != spacedim:
        #     for ax in range(spacedim):
        #         if ax not in self.vectorder:
        #             affines[:, ax, ax] = 1
        # Instead of the above, fix rank deficiency empirically, as it may also
        # arise when the displacement along one dimension is infinitesimally
        # small, albeit the dimension per se is defined.
        def fix_rank_deficiency(a, columns, cond=1e-5):
            for col in np.asarray(columns, dtype=np.int):
                diagonals = a[:, col]
                a[:, col] = np.where(diagonals < cond, 1, diagonals)
            return a

        if affines.shape[-1] == 12:
            affines = fix_rank_deficiency(affines, [0, 4, 8])
            # Note that the C-function uses LAPACK routines and exports the
            # matrix elements in Fortran order
            affines = affines.reshape((-1, 3, 4), order="F")

        elif affines.shape[-1] == 6:
            affines = fix_rank_deficiency(affines, [0, 3])
            affines = affines.reshape((-1, 2, 3), order="F")

        else:
            raise AssertionError(
                f"Unexpected number of affine components: {affines.shape[-1]}")

        # The resultant affines are independent of the transformation's vector
        # order, and they transform xy(z) coordinates. This is in keeping with
        # the fact that DisplacementFields with a non-standard vector order
        # also transform input coordinates in the xy(z) layout. The vector
        # order only refers to the storage order of the vectors within the
        # transformation.

        return affines

    def inverse(self, domain=None):
        """
        Returns the inverse transformation, which is by default a dense
        displacement field, defined on a domain whose shape is specified by
        the transformation's dense_shape attribute.

        If necessary, the inverse transformation can be reduced to an
        RbfDisplacementField by evaluating the DisplacementField on a
        non-compact domain, and casting the result as RbfDisplacementField.

        While a sparse inverse transformation may only be fully accurate at the
        support points, as opposed to a dense DisplacementField, it will have a
        non-zero mapping outside the bounding box of the support points.

        """
        # If the input domain is not a compact domain,
        # create a suitable one that cointains all support points
        if getattr(domain, "iscompact", False):
            dense = domain
        else:
            name = self.domain.name + "_dense"
            dense = Domain(
                self.dense_shape, *self.domain.chain,
                offset=self.domain.offset[1:], name=name,
                storage=self.domain.storage,
                instance_mem_limit=self.domain.instance_mem_limit,
                n_threads=self.domain.threads, **tu.rcopy(self.domain.kwargs))

        # Call the parent-class method to create a TxDisplacementField instance
        field = self.std().absolute().field(dense)
        inv_tx = TxDisplacementField(field).inverse()

        # If the input domain was non-compact, return a TxRbfDisplacementField
        # by res
        if not getattr(domain, "iscompact", True):
            field = inv_tx.field(domain)
            inv_tx = TxRbfDisplacementField(field, mode=self.mode)

        return inv_tx

    def jacobian(self, coords=None):
        # TODO: account for absolute/relative + vector order!

        # Obtain SciPy Rbf interpolator instance, and use its private functions
        # instead of implementing everything here. There is a risk that these
        # functions may be changed in future versions of SciPy. Should that
        # happen, this function needs to be changed as well.
        rbf = self._dummy_interpolator()

        # Obtain input coordinates in the pseudo-voxel space where the nodes
        # of the RbfInterpolator are defined.
        if coords is None:
            vc = self.domain.get_voxel_coordinates()
            vc = self.domain.offset.map(vc)
        else:
            vc = self.domain.chain.inverse().map(coords)

        # Calculate distance matrix (n_points, n_nodes)
        distance_matrix = rbf._call_norm(vc.T, rbf.xi)
        g = rbf._function(distance_matrix)
        # Calculate the transformation Jacobian for the vector field
        # (stacked scalar field parameters)
        jac = np.einsum("ki,ij->kj", g, np.linalg.inv(rbf.A))
        # Repeat jacobian for all displacement dimensions
        d = self.vectdim
        return np.kron(np.eye(d, d), jac[:, np.newaxis, :])

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~ PRIVATE METHODS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

    def _dummy_interpolator(self):
        ip = self.interpolator
        return Rbf(*ip.basis, np.zeros(self.domain.numel), **ip.ipkwargs)

    @staticmethod
    def _d_multiquadric(r, epsilon):
        return r / (epsilon * np.sqrt(epsilon ** 2. + r ** 2.))

    @staticmethod
    def _dd_multiquadric(r, epsilon):
        return epsilon ** 2. / ((epsilon ** 2. + r ** 2.) ** (3. / 2.))

    @staticmethod
    def _d_inverse(r, epsilon):
        return epsilon * r / ((epsilon ** 2. + r ** 2.) ** (3. / 2.))

    @staticmethod
    def _dd_inverse(r, epsilon):
        return epsilon * (epsilon ** 2. - 2. * r ** 2.) / \
               ((epsilon ** 2. + r ** 2.) ** (5./2.))

    @staticmethod
    def _d_gaussian(r, epsilon):
        return -2. * r / (epsilon ** 2.) * np.exp(-((r / epsilon) ** 2.))

    @staticmethod
    def _dd_gaussian(r, epsilon):
        return 2. * (2. * r ** 2. - epsilon ** 2.) / (epsilon ** 4.) * \
               np.exp(-((r / epsilon) ** 2.))

    @staticmethod
    def _d_linear(r, epsilon=None):
        return 1.

    @staticmethod
    def _dd_linear(r, epsilon=None):
        return 0.

    @staticmethod
    def _d_cubic(r, epsilon=None):
        return 3. * (r ** 2.)

    @staticmethod
    def _dd_cubic(r, epsilon=None):
        return 6. * r

    @staticmethod
    def _d_quintic(r, epsilon=None):
        return 5. * (r ** 4.)

    @staticmethod
    def _dd_quintic(r, epsilon=None):
        return 20. * (r ** 3.)

    @staticmethod
    def _d_thin_plate(r, epsilon=None):
        return np.where(r == 0, 0, r * (2. * np.log(r) + 1.))

    @staticmethod
    def _dd_thin_plate(r, epsilon=None):
        return np.where(r == 0, 0, 2. * np.log(r) + 3.)

    def _d_current_model(self, r):
        """ Returns the first derivative of the radial basis function that is
        being used by the transformation object. """
        model = self.interpolator.model
        if isinstance(model, str):
            func = getattr(self, "_d_{}".format(model.lower()))
        elif hasattr(model, "__call__"):
            func = getattr(self, "_d_{}".format(model.__name__))
        else:
            raise ValueError("Invalid radial basis function definition.")
        return func(r, self.interpolator.epsilon)

    def _dd_current_model(self, r):
        """ Returns the second derivative of the radial basis function that is
        being used by the transformation object. """
        model = self.interpolator.model
        if isinstance(model, str):
            func = getattr(self, "_dd_{}".format(model.lower()))
        elif hasattr(model, "__call__"):
            func = getattr(self, "_dd_{}".format(model.__name__))
        else:
            raise ValueError("Invalid radial basis function definition.")
        return func(r, self.interpolator.epsilon)

    def _dump(self):
        objdump = super(TxRbfDisplacementField, self)._dump()
        dense_shape = [int(dim) for dim in self.dense_shape]
        objdump["metaparameters"]["dense_shape"] = dense_shape
        return objdump


if __name__ == "__main__":  # pragma: no cover
    print("""This module is not intended for execution.""")
